from typing import Any
from typing import Literal
from typing import overload

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import GtkSource

class GajimBuilder:
    def __init__(
        self,
        filename: str | None = None,
        instance: Any = None,
        widgets: list[str] | None = None,
        domain: str | None = None,
        gettext_: Any | None = None,
    ) -> None: ...

class Builder(Gtk.Builder): ...

class ActivityChangeTimezoneBuilder(Builder):
    change_timezone_page: Gtk.Box
    old_timezone_label: Gtk.Label
    new_timezone_label: Gtk.Label
    update_button: Gtk.Button
    ignore_button: Gtk.Button
    ask_button: Gtk.CheckButton

class ActivityDefaultBuilder(Builder):
    default_page: Gtk.Box

class ActivityGajimUpdateBuilder(Builder):
    gajim_update_page: Gtk.Box
    update_permission_box: Gtk.Box
    update_permission_title_label: Gtk.Label
    update_permission_enable_button: Gtk.Button
    update_gajim_box: Gtk.Box
    update_gajim_title_label: Gtk.Label
    update_gajim_text_label: Gtk.Label
    update_gajim_button: Gtk.Button
    update_plugins_box: Gtk.Box
    update_plugins_title_label: Gtk.Label
    update_plugins_text_label: Gtk.Label
    update_plugins_automatically_checkbox: Gtk.CheckButton
    update_plugins_show_button: Gtk.Button
    update_plugins_button: Gtk.Button
    update_plugins_success_box: Gtk.Box
    update_plugins_success_title_label: Gtk.Label
    update_plugins_success_text_label: Gtk.Label
    update_plugins_success_button: Gtk.Button

class ActivityMucInvitationBuilder(Builder):
    muc_invitation_page: Gtk.Box
    muc_invitation_box: Gtk.Box
    muc_invitation_declined_box: Gtk.Box
    invitation_image: Gtk.Image
    invitation_title_label: Gtk.Label
    invitation_text_label: Gtk.Label

class ActivitySubscriptionBuilder(Builder):
    subscription_page: Gtk.Box
    subscription_image: Gtk.Image
    subscription_title_label: Gtk.Label
    subscription_text_label: Gtk.Label
    subscribe_box: Gtk.Box
    subscribe_deny_button: Gtk.Button
    subscribe_accept_button: Gtk.Button
    subscribe_menu_button: Gtk.MenuButton
    unsubscribed_box: Gtk.Box
    unsubscribed_remove_button: Gtk.Button

class AdvancedConfigurationBuilder(Builder):
    box: Gtk.Box
    search_entry: Gtk.SearchEntry
    advanced_treeview: Gtk.TreeView
    treeview_selection: Gtk.TreeSelection
    description: Gtk.Label
    reset_button: Gtk.Button

class AssistantBuilder(Builder):
    main_box: Gtk.Box
    content_area: Gtk.Box
    stack: Gtk.Stack
    action_area: Gtk.Box

class CallWindowBuilder(Builder):
    adjustment1: Gtk.Adjustment
    adjustment2: Gtk.Adjustment
    dtmf_popover: Gtk.Popover
    grid1: Gtk.Grid
    button_1: Gtk.Button
    button_2: Gtk.Button
    button_3: Gtk.Button
    button_4: Gtk.Button
    button_5: Gtk.Button
    button_6: Gtk.Button
    button_7: Gtk.Button
    button_8: Gtk.Button
    button_9: Gtk.Button
    button_star: Gtk.Button
    button_0: Gtk.Button
    button_pound: Gtk.Button
    av_box: Gtk.Box
    video_box: Gtk.Box
    outgoing_viewport: Gtk.Viewport
    incoming_viewport: Gtk.Viewport
    avatar_image: Gtk.Image
    jingle_audio_state: Gtk.Image
    jingle_connection_state: Gtk.Label
    jingle_connection_spinner: Adw.Spinner
    answer_video_button: Gtk.Button
    av_cam_button: Gtk.Button
    av_cam_image: Gtk.Image
    end_call_button: Gtk.Button
    audio_buttons_box: Gtk.Box
    mic_hscale: Gtk.VolumeButton
    volumebutton_plus_button2: Gtk.Button
    volumebutton_minus_button2: Gtk.Button
    sound_hscale: Gtk.VolumeButton
    volumebutton_plus_button1: Gtk.Button
    volumebutton_minus_button1: Gtk.Button
    dtmf_button: Gtk.MenuButton

class ChatControlBuilder(Builder):
    control_box: Gtk.Box
    conv_view_paned: Gtk.Paned
    conv_view_overlay: Gtk.Overlay

class ChatListRowBuilder(Builder):
    mainbox: Gtk.Box
    account_identifier: Gtk.Box
    avatar_image: Gtk.Image
    connection_icon: Gtk.Image
    group_chat_indicator: Gtk.Image
    name_label: Gtk.Label
    chatstate_image: Gtk.Image
    mute_image: Gtk.Image
    timestamp_label: Gtk.Label
    nick_label: Gtk.Label
    message_icon: Gtk.Image
    message_label: Gtk.Label
    unread_label: Gtk.Label
    revealer: Gtk.Revealer
    close_button: Gtk.Button

class ContactInfoBuilder(Builder):
    groups_model: Gtk.ListStore
    main_stack: Gtk.Stack
    vcard_scrolled: Gtk.ScrolledWindow
    avatar_image: Gtk.Image
    contact_name_controls_box: Gtk.Box
    contact_jid_label: Gtk.Label
    group_chat_info: Adw.PreferencesGroup
    role_info: Adw.ActionRow
    affiliation_info: Adw.ActionRow
    vcard_box: Gtk.Box
    settings_scrolled: Gtk.ScrolledWindow
    settings_box: Gtk.Box
    subscription_box: Gtk.Box
    subscription_listbox: Gtk.ListBox
    from_subscription_switch: Gtk.Switch
    to_subscription_stack: Gtk.Stack
    request_stack: Gtk.Stack
    to_subscription_button: Gtk.Button
    remove_history_button: Gtk.Button
    encryption_scrolled: Gtk.ScrolledWindow
    encryption_box: Adw.Clamp
    groups_page_stack: Gtk.Stack
    groups_treeview: Gtk.TreeView
    tree_selection: Gtk.TreeSelection
    toggle_renderer: Gtk.CellRendererToggle
    text_renderer: Gtk.CellRendererText
    group_add_button: Gtk.Button
    group_remove_button: Gtk.Button
    notes_page_stack: Gtk.Stack
    scrolledwindow_annotation: Gtk.ScrolledWindow
    textview_annotation: Gtk.TextView
    devices_stack: Gtk.Stack
    devices_page: Adw.PreferencesPage

class DbMigrationBuilder(Builder):
    box: Gtk.Box
    stack: Gtk.Stack
    status_label: Gtk.Label
    version_label: Gtk.Label
    error_label: Gtk.Label
    error_view: Gtk.TextView
    error_copy_button: Gtk.Button
    error_close_button: Gtk.Button
    success_close_button: Gtk.Button

class DebugConsoleBuilder(Builder):
    popover: Gtk.Popover
    stanza_presets_listbox: Gtk.ListBox
    stack: Gtk.Stack
    log_view: GtkSource.View
    paned: Gtk.Paned
    search_revealer: Gtk.Revealer
    search_entry: Gtk.SearchEntry
    search_forward: Gtk.Button
    search_backward: Gtk.Button
    search_results_label: Gtk.Label
    scrolled: Gtk.ScrolledWindow
    protocol_view: GtkSource.View
    jump_to_end_button: Gtk.Button
    scrolled_input: Gtk.ScrolledWindow
    input_entry: GtkSource.View
    actionbox: Gtk.Box
    filter_options_button: Gtk.Button
    clear_button: Gtk.Button
    edit_toggle: Gtk.ToggleButton
    paste: Gtk.Button
    menubutton: Gtk.MenuButton
    account_label: Gtk.Label
    send: Gtk.Button
    header_box: Gtk.Box
    search_toggle: Gtk.ToggleButton

class ExceptionDialogBuilder(Builder):
    exception_box: Gtk.Box
    infobar: Gtk.Revealer
    exception_view: Gtk.TextView
    user_feedback_box: Gtk.Box
    user_feedback_entry: Gtk.Entry
    close_button: Gtk.Button
    report_spinner: Adw.Spinner
    report_button: Gtk.Button

class FileTransferBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_description: Gtk.Label
    file_name: Gtk.Label
    file_size: Gtk.Label
    progress_bar: Gtk.ProgressBar
    transfer_progress: Gtk.Label
    cancel_button: Gtk.Button

class FileTransferJingleBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_action: Gtk.Label
    file_name: Gtk.Label
    file_description: Gtk.Label
    file_size: Gtk.Label
    action_stack: Gtk.Stack
    accept_file_request: Gtk.Button
    reject_file_request: Gtk.Button
    open_folder: Gtk.Button
    open_file: Gtk.Button
    error_show_transfers: Gtk.Button
    error_label: Gtk.Label
    retry_bad_hash: Gtk.Button
    rejected_show_transfers: Gtk.Button
    progress_label: Gtk.Label
    progress_bar: Gtk.ProgressBar
    cancel_transfer: Gtk.Button

class FileTransferSelectorBuilder(Builder):
    file_box: Gtk.Box
    preview_image_box: Gtk.Box
    preview_image: Gtk.Image
    file_name_label: Gtk.Label
    file_size_label: Gtk.Label
    warning_label: Gtk.Label
    remove_file_button: Gtk.Button
    stack: Gtk.Stack
    box: Gtk.Box
    listbox: Gtk.ListBox
    resource_box: Gtk.Box
    resource_instructions: Gtk.Label

class GroupchatAffiliationBuilder(Builder):
    affiliation_store: Gtk.ListStore
    combo_store: Gtk.ListStore
    main_box: Gtk.Box
    affiliation_scrolled: Gtk.ScrolledWindow
    affiliation_treeview: Gtk.TreeView
    affiliation_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reserved_name_column: Gtk.TreeViewColumn
    reserved_name_renderer: Gtk.CellRendererText
    affiliation_renderer: Gtk.CellRendererCombo
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button

class GroupchatBlocksBuilder(Builder):
    main: Gtk.Box
    top_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    remove_button: Gtk.Button
    scrolled_box: Gtk.Box
    scrolled: Gtk.ScrolledWindow
    column_view: Gtk.ColumnView
    nickname_col: Gtk.ColumnViewColumn
    id_col: Gtk.ColumnViewColumn

class GroupchatConfigBuilder(Builder):
    stack: Gtk.Stack
    loading_box: Gtk.Box
    config_box: Gtk.Box
    error_box: Gtk.Box
    error_image: Gtk.Image
    error_label: Gtk.Label

class GroupchatDetailsBuilder(Builder):
    main_stack: Gtk.Stack
    info_container: Gtk.Box
    settings_box: Gtk.Box
    encryption_scrolled: Gtk.ScrolledWindow
    encryption_box: Adw.Clamp
    blocks_box: Gtk.Box
    manage_box: Gtk.Box
    affiliation_box: Gtk.Box
    outcasts_box: Gtk.Box
    configuration_box: Gtk.Box

class GroupchatInfoScrolledBuilder(Builder):
    info_clamp: Adw.Clamp
    avatar_image: Gtk.Image
    name_box: Gtk.Box
    info_listbox: Gtk.ListBox
    address_row: Adw.ActionRow
    address_copy_button: Gtk.Button
    description_row: Adw.ActionRow
    subject_row: Adw.ActionRow
    author_row: Adw.ActionRow
    users_row: Adw.ActionRow
    contact_row: Adw.ActionRow
    contact_box: Gtk.Box
    logs_row: Adw.ActionRow
    logs: Gtk.LinkButton
    lang_row: Adw.ActionRow
    features_group: Adw.PreferencesGroup
    features_listbox: Gtk.ListBox

class GroupchatInviterBuilder(Builder):
    account_store: Gtk.ListStore
    invite_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    scrolledwindow: Gtk.ScrolledWindow
    contacts_listbox: Gtk.ListBox
    contacts_placeholder: Gtk.Box
    invitees_scrolled: Gtk.ScrolledWindow

class GroupchatManageBuilder(Builder):
    scrolled: Gtk.ScrolledWindow
    stack: Gtk.Stack
    avatar_overlay: Gtk.Overlay
    avatar_button_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    manage_save_button: Gtk.Button
    contact_addresses_listbox: Gtk.ListBox
    muc_name_entry_row: Adw.EntryRow
    muc_description_entry_row: Adw.EntryRow
    destroy_muc_button: Gtk.Button
    subject_change_button: Gtk.Button
    subject_textview: Gtk.TextView
    avatar_selector_grid: Gtk.Grid
    avatar_cancel_button: Gtk.Button
    avatar_update_button: Gtk.Button
    destroy_reason_entry: Gtk.Entry
    destroy_alternate_entry: Gtk.Entry
    destroy_cancel_button: Gtk.Button
    destroy_button: Gtk.Button

class GroupchatOutcastBuilder(Builder):
    info_popover: Gtk.Popover
    outcast_store: Gtk.ListStore
    main_box: Gtk.Box
    outcast_scrolled: Gtk.ScrolledWindow
    outcast_treeview: Gtk.TreeView
    outcast_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reason_renderer: Gtk.CellRendererText
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button
    info_button: Gtk.MenuButton

class GroupchatRosterBuilder(Builder):
    box: Gtk.Box
    participants_count_label: Gtk.Label
    search_entry: Gtk.SearchEntry
    scrolled: Gtk.ScrolledWindow

class GroupchatRosterTooltipBuilder(Builder):
    tooltip_grid: Gtk.Grid
    avatar: Gtk.Image
    jid: Gtk.Label
    nick: Gtk.Label
    fillelement: Gtk.Label
    status: Gtk.Label
    affiliation: Gtk.Label
    hats_box: Gtk.Box

class GroupchatStateBuilder(Builder):
    groupchat_state: Gtk.Stack
    join_button: Gtk.Button
    joining_spinner: Adw.Spinner
    abort_join_button: Gtk.Button
    mam_sync_spinner: Adw.Spinner
    mam_error_label: Gtk.Label
    close_button: Gtk.Button

class HistoryExportBuilder(Builder):
    select_account_box: Gtk.Box
    settings_grid: Gtk.Grid

class ManageSoundsBuilder(Builder):
    manage_sounds: Gtk.Box
    sounds_treeview: Gtk.TreeView
    toggle_cell_renderer: Gtk.CellRendererToggle
    sound_buttons_box: Gtk.Box
    clear_sound_button: Gtk.Button
    play_sound_button: Gtk.Button

class MessageActionsBoxBuilder(Builder):
    box: Gtk.Box
    chat_state_box: Gtk.Box
    link_previews_box: Gtk.Box
    reply_box: Gtk.Box
    state_box: Gtk.Box
    state_box_image: Gtk.Image
    state_box_label: Gtk.Label
    visitor_menu_button: Gtk.MenuButton
    edit_box: Gtk.Box
    edit_box_image: Gtk.Image
    cancel_correction_button: Gtk.Button
    action_box: Gtk.Box
    emoticons_button: Gtk.MenuButton
    formattings_button: Gtk.MenuButton
    input_wrapper: Gtk.Overlay
    input_scrolled: Gtk.ScrolledWindow
    input_overlay: Gtk.Box
    input_overlay_label: Gtk.Label
    send_message_button: Gtk.Button
    sendfile_button: Gtk.Button
    encryption_menu_button: Gtk.MenuButton
    encryption_image: Gtk.Image
    encryption_details_button: Gtk.Button
    encryption_details_image: Gtk.Image
    visitor_popover: Gtk.Popover
    request_voice_button: Gtk.Button

class OmemoTrustManagerBuilder(Builder):
    qr_code_popover: Gtk.Popover
    comparing_instructions: Gtk.Label
    our_fingerprint_2: Gtk.Label
    qr_code_image: Gtk.Image
    stack: Gtk.Stack
    our_fingerprint_row: Adw.ActionRow
    copy_button: Gtk.Button
    qr_menu_button: Gtk.MenuButton
    manage_trust_button: Gtk.Button
    list_heading: Gtk.Label
    list_heading_box: Gtk.Box
    show_inactive_switch: Gtk.Switch
    clear_devices_button: Gtk.Button
    list: Gtk.ListBox
    undecided_placeholder: Gtk.Label

class PasswordDialogBuilder(Builder):
    pass_box: Gtk.Box
    header: Gtk.Label
    message_label: Gtk.Label
    pass_entry: Gtk.PasswordEntry
    save_pass_checkbutton: Gtk.CheckButton
    keyring_hint: Gtk.Label
    cancel_button: Gtk.Button
    ok_button: Gtk.Button

class PepConfigBuilder(Builder):
    stack: Gtk.Stack
    overview_box: Gtk.Box
    services_treeview: Gtk.TreeView
    reload_button: Gtk.Button
    delete_button: Gtk.Button
    configure_button: Gtk.Button
    show_content_button: Gtk.Button
    items_box: Gtk.Box
    items_label: Gtk.Label
    items_view: GtkSource.View
    items_back_button: Gtk.Button
    config_box: Gtk.Box
    form_label: Gtk.Label
    form_box: Gtk.Box
    config_back_button: Gtk.Button
    save_button: Gtk.Button

class ProfileBuilder(Builder):
    privacy_popover: Gtk.Popover
    avatar_nick_access: Gtk.Switch
    vcard_access: Gtk.Switch
    avatar_nick_access_label: Gtk.Label
    vcard_access_label: Gtk.Label
    profile_stack: Gtk.Stack
    spinner: Adw.Spinner
    scrolled: Gtk.ScrolledWindow
    profile_box: Gtk.Box
    avatar_overlay: Gtk.Overlay
    avatar_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    nickname_entry: Gtk.Entry
    nickname_label: Gtk.Label
    cancel_button: Gtk.Button
    add_entry_button: Gtk.MenuButton
    privacy_button: Gtk.MenuButton
    save_button: Gtk.Button
    edit_button: Gtk.Button
    avatar_selector_box: Gtk.Box
    avatar_cancel: Gtk.Button
    avatar_update_button: Gtk.Button
    error_label: Gtk.Label
    error_title_label: Gtk.Label
    back_button: Gtk.Button

class QuitDialogBuilder(Builder):
    box: Gtk.Box
    remember_checkbutton: Gtk.CheckButton
    hide_button: Gtk.Button
    minimize_button: Gtk.Button
    quit_button: Gtk.Button

class RosterItemExchangeBuilder(Builder):
    roster_item_exchange: Gtk.Box
    type_label: Gtk.Label
    body_scrolledwindow: Gtk.ScrolledWindow
    body_textview: Gtk.TextView
    items_list_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    cancel_button: Gtk.Button
    accept_button: Gtk.Button

class SearchViewBuilder(Builder):
    calendar_popover: Gtk.Popover
    calendar: Gtk.Calendar
    first_day_button: Gtk.Button
    previous_day_button: Gtk.Button
    next_day_button: Gtk.Button
    last_day_button: Gtk.Button
    search_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    calendar_button: Gtk.MenuButton
    close_button: Gtk.Button
    search_checkbutton: Gtk.CheckButton
    search_filters_box: Gtk.Box
    results_scrolled: Gtk.ScrolledWindow
    results_listbox: Gtk.ListBox
    status_page: Adw.StatusPage
    filter_date_before_popover: Gtk.Popover
    filter_date_before_calendar: Gtk.Calendar
    filter_date_before_reset_button: Gtk.Button
    filter_date_after_popover: Gtk.Popover
    filter_date_after_calendar: Gtk.Calendar
    filter_date_after_reset_button: Gtk.Button
    search_filters_grid: Gtk.Grid
    filter_from_desc_label: Gtk.Label
    filter_from_entry: Gtk.Entry
    filter_before_button: Gtk.MenuButton
    filter_after_button: Gtk.MenuButton
    filter_before_label: Gtk.Label
    filter_after_label: Gtk.Label
    header_box: Gtk.Box
    header_name_label: Gtk.Label
    header_date_label: Gtk.Label
    result_row_grid: Gtk.Grid
    row_avatar: Gtk.Image
    row_time_label: Gtk.Label
    row_name_label: Gtk.Label

class ServiceDiscoveryWindowBuilder(Builder):
    service_discovery: Gtk.Box
    banner_agent_icon: Gtk.Image
    banner_agent_header: Gtk.Label
    banner_agent_subheader: Gtk.Label
    address_box: Gtk.Box
    address_comboboxtext: Gtk.ComboBoxText
    browse_button: Gtk.Button
    services_progressbar: Gtk.ProgressBar
    services_scrollwin: Gtk.ScrolledWindow
    services_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    action_buttonbox: Gtk.Box

class SslErrorDialogBuilder(Builder):
    ssl_error_box: Gtk.Box
    intro_text: Gtk.Label
    ssl_error: Gtk.Label
    add_certificate_checkbutton: Gtk.CheckButton
    view_cert_button: Gtk.Button
    connect_button: Gtk.Button

class StartChatDialogBuilder(Builder):
    stack: Gtk.Stack
    infobar: Gtk.Revealer
    infobar_close_button: Gtk.Button
    box: Gtk.Box
    controls_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    global_search_toggle: Gtk.ToggleButton
    settings_menu: Gtk.MenuButton
    search_error_box: Gtk.Box
    search_error_label: Gtk.Label
    list_stack: Gtk.Stack
    contact_scrolled: Gtk.ScrolledWindow
    no_contacts_placeholder: Gtk.Box
    global_scrolled: Gtk.ScrolledWindow
    global_search_placeholder_stack: Gtk.Stack
    global_search_placeholder_hints: Gtk.Box
    global_search_results_label: Gtk.Label
    spinner: Adw.Spinner
    error_label: Gtk.Label
    error_back_button: Gtk.Button
    info_box: Gtk.Box
    info_back_button: Gtk.Button
    join_box: Gtk.Box
    join_button: Gtk.Button
    account_view: Gtk.TreeView
    icon_pixbuf: Gtk.CellRendererPixbuf
    account_text: Gtk.CellRendererText
    account_back_button: Gtk.Button
    account_select_button: Gtk.Button

class ThemesWindowBuilder(Builder):
    option_popover: Gtk.Popover
    choose_option_listbox: Gtk.ListBox
    placeholder: Gtk.Box
    theme_store: Gtk.ListStore
    theme_grid: Gtk.Grid
    theme_treeview: Gtk.TreeView
    theme_treeview_selection: Gtk.TreeSelection
    theme_name_cell_renderer: Gtk.CellRendererText
    option_listbox: Gtk.ListBox
    add_option_button: Gtk.MenuButton
    add_theme_button: Gtk.Button
    remove_theme_button: Gtk.Button

class VideoPreviewBuilder(Builder):
    video_preview_box: Gtk.Box
    video_source_label: Gtk.Label
    video_preview_placeholder: Gtk.Box

class VoiceMessageRecorderBuilder(Builder):
    popover: Gtk.Popover
    box: Gtk.Box
    error_label: Gtk.Label
    progression_box: Gtk.Box
    time_label: Gtk.Label
    visualization_box: Gtk.Box
    audio_player_box: Gtk.Box
    record_control_box: Gtk.Box
    cancel_button: Gtk.Button
    record_toggle_button: Gtk.Button
    record_toggle_button_image: Gtk.Image
    send_button: Gtk.Button

class WorkspaceDialogBuilder(Builder):
    box: Gtk.Box
    preview: Gtk.Image
    entry: Gtk.Entry
    remove_workspace_button: Gtk.Button
    image_switch: Gtk.Switch
    style_stack: Gtk.Stack
    color_dialog_button: Gtk.ColorDialogButton
    image_box: Gtk.Box
    cancel_button: Gtk.Button
    save_button: Gtk.Button

@overload
def get_builder(
    file_name: Literal["activity_change_timezone.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityChangeTimezoneBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_default.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityDefaultBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_gajim_update.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityGajimUpdateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_muc_invitation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivityMucInvitationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["activity_subscription.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ActivitySubscriptionBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["advanced_configuration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> AdvancedConfigurationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["assistant.ui"], instance: Any = None, widgets: list[str] = ...
) -> AssistantBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["call_window.ui"], instance: Any = None, widgets: list[str] = ...
) -> CallWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_control.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatControlBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_list_row.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatListRowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["contact_info.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ContactInfoBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["db_migration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DbMigrationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["debug_console.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DebugConsoleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["exception_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ExceptionDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_jingle.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferJingleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_selector.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferSelectorBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_affiliation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatAffiliationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_blocks.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatBlocksBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_config.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_details.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatDetailsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_info_scrolled.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInfoScrolledBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_inviter.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInviterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_manage.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatManageBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_outcast.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatOutcastBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster_tooltip.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterTooltipBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_state.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatStateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["history_export.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> HistoryExportBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["manage_sounds.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ManageSoundsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["message_actions_box.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> MessageActionsBoxBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["omemo_trust_manager.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> OmemoTrustManagerBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["password_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> PasswordDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["pep_config.ui"], instance: Any = None, widgets: list[str] = ...
) -> PepConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["profile.ui"], instance: Any = None, widgets: list[str] = ...
) -> ProfileBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["quit_dialog.ui"], instance: Any = None, widgets: list[str] = ...
) -> QuitDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["roster_item_exchange.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> RosterItemExchangeBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["search_view.ui"], instance: Any = None, widgets: list[str] = ...
) -> SearchViewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["service_discovery_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ServiceDiscoveryWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["ssl_error_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> SslErrorDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["start_chat_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> StartChatDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["themes_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ThemesWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["video_preview.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VideoPreviewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["voice_message_recorder.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VoiceMessageRecorderBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["workspace_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> WorkspaceDialogBuilder: ...  # noqa
def get_builder(
    file_name: str, instance: Any = None, widgets: list[str] = ...
) -> Builder: ...



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk
from nbxmpp.modules.security_labels import SecurityLabel
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import SimpleClientState
from gajim.common.events import SecCatalogReceived
from gajim.common.i18n import _
from gajim.common.types import ChatContactT

from gajim.gtk.dropdown import GajimDropDown


class SecurityLabelSelector(GajimDropDown[str]):
    __gtype_name__ = "SecurityLabelSelector"

    def __init__(self) -> None:
        super().__init__(fixed_width=15)
        self.set_visible(False)

        self._client: Client | None = None
        self._contact: ChatContactT | None = None

        self.set_valign(Gtk.Align.CENTER)
        self.set_tooltip_text(_("Select a security label for your message…"))

        app.ged.register_event_handler(
            "sec-catalog-received", ged.GUI1, self._sec_labels_received
        )

    def switch_contact(self, contact: ChatContactT) -> None:
        self.clear()

        self._client = app.get_client(contact.account)
        self._contact = contact

        app.settings.connect_signal(
            "enable_security_labels",
            self._on_setting_changed,
            account=self._contact.account,
        )
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._update_data()

    def do_unroot(self) -> None:
        self.clear()
        app.ged.remove_event_handler(
            "sec-catalog-received", ged.GUI1, self._sec_labels_received
        )
        super().do_unroot()
        app.check_finalize(self)

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        if state == SimpleClientState.CONNECTED:
            self._update_data()

    def _on_setting_changed(
        self, state: bool, _name: str, _account: str | None, _jid: JID | None
    ) -> None:
        if state:
            self._update_data()
        else:
            self.set_visible(False)
            self.set_data(None)

    def _sec_labels_received(self, event: SecCatalogReceived) -> None:
        if self._contact is None or self._contact.account != event.account:
            return

        if event.jid != self._contact.jid:
            return

        self._update_data()

    def _update_data(self) -> None:
        assert self._contact is not None
        assert self._client is not None

        if not app.settings.get_account_setting(
            self._contact.account, "enable_security_labels"
        ):
            self.set_visible(False)
            return

        catalog = self._client.get_module("SecLabels").get_catalog(self._contact.jid)

        self.set_data(None)

        if catalog is not None:
            selectors = catalog.get_label_names()
            self.set_data(selectors)
            self.select_key(catalog.default)

        self.set_visible(True)

    def get_seclabel(self) -> SecurityLabel | None:
        assert self._contact is not None
        if not app.settings.get_account_setting(
            self._contact.account, "enable_security_labels"
        ):
            return

        selector = self.get_selected_key()
        if selector is None:
            return

        assert self._client is not None
        catalog = self._client.get_module("SecLabels").get_catalog(self._contact.jid)
        if catalog is None:
            return None
        return catalog.labels[selector]

    def set_seclabel(self, label_hash: str) -> None:
        assert self._contact is not None
        assert self._client is not None

        catalog = self._client.get_module("SecLabels").get_catalog(self._contact.jid)
        if catalog is None:
            return None

        for selector, label in catalog.labels.items():
            if label is None:
                continue
            if label.get_label_hash() == label_hash:
                self.select_key(selector)

    def clear(self) -> None:
        app.settings.disconnect_signals(self)
        if self._client is not None:
            self._client.disconnect_all_from_obj(self)

        self._client = None
        self._contact = None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import datetime
import logging
import math
import sys
from collections import deque
from collections.abc import Callable
from pathlib import Path

from gi.repository import GObject

from gajim.common import app
from gajim.common import configpaths
from gajim.common.i18n import _

try:
    from gi.repository import Gst
except Exception:
    if TYPE_CHECKING:
        from gi.repository import Gst

log = logging.getLogger("gajim.gtk.voice_message_recorder")


GST_ERROR_ON_START = 0
GST_ERROR_ON_RECORDING = 1
GST_ERROR_ON_STOP = 2
GST_ERROR_ON_MERGING = 3


class VoiceMessageRecorder(GObject.GObject):
    __gtype_name__ = "VoiceMessageRecorder"

    def __init__(self, error_callback: Callable[[int, str], None]) -> None:
        GObject.GObject.__init__(self)
        if not app.is_installed("GST"):
            raise Exception("Missing GStreamer dependency")

        self._available = True
        self._error_callback = error_callback

        # Recording state
        self._new_recording = True
        self._output_file_counter = 0
        self._output_file_valid = False
        self._output_files_invalid: list[int] = []
        self._num_samples = 80
        self._num_samples_buffer = 20
        self._samples: deque[float] = deque(
            [0.0] * self._num_samples, maxlen=self._num_samples
        )
        self._samples_buffer: deque[float] = deque(
            [0.0] * self._num_samples_buffer, maxlen=self._num_samples_buffer
        )
        self._buffer_level = 0
        self._rec_time = {"start": 0, "total": 0}

        app.settings.connect_signal(
            "audio_input_device", self._on_audio_input_device_changed
        )

        # Voice message storage location
        self._filetype = "m4a"
        timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        self._file_name = f"voice-message-{timestamp}.{self._filetype}"
        self._file_path = configpaths.get_temp_dir() / self._file_name

        self._init_pipeline()

    def _init_pipeline(self) -> None:
        log.debug("Setting up pipeline")

        self._pipeline = Gst.Pipeline.new("pipeline")
        self._queue = Gst.ElementFactory.make("queue")
        audioconvert = Gst.ElementFactory.make("audioconvert")
        audioresample = Gst.ElementFactory.make("audioresample")
        audiolevel = Gst.ElementFactory.make("level")
        opusenc = Gst.ElementFactory.make("opusenc")
        mp4mux = Gst.ElementFactory.make("mp4mux")
        self._filesink = Gst.ElementFactory.make("filesink")

        pipeline_elements = [
            self._pipeline,
            self._queue,
            audioconvert,
            audioresample,
            audiolevel,
            opusenc,
            mp4mux,
            self._filesink,
        ]

        if any(element is None for element in pipeline_elements):
            raise Exception("Error while setting up pipeline")

        assert self._pipeline is not None
        assert self._queue is not None
        assert audioconvert is not None
        assert audioresample is not None
        assert audiolevel is not None
        assert opusenc is not None
        assert mp4mux is not None
        assert self._filesink is not None

        opusenc.set_property("audio-type", "voice")
        audiolevel.set_property("message", True)
        self._filesink.set_property("location", str(self._file_path))
        self._filesink.set_property("async", False)

        self._pipeline.add(self._queue)
        self._pipeline.add(audioconvert)
        self._pipeline.add(audioresample)
        self._pipeline.add(audiolevel)
        self._pipeline.add(opusenc)
        self._pipeline.add(mp4mux)
        self._pipeline.add(self._filesink)

        self._queue.link(audioconvert)
        audioconvert.link(audioresample)
        audioresample.link(audiolevel)
        audiolevel.link(opusenc)
        opusenc.link(mp4mux)
        mp4mux.link(self._filesink)

        self._clock: Gst.Clock | None = None

        self._bus = self._pipeline.get_bus()
        self._bus.add_signal_watch()
        self._id = self._bus.connect("message", self._on_gst_message)

        self._audiosrc = self._create_new_audio_src()
        if self._audiosrc is not None:
            self._pipeline.add(self._audiosrc)
            self._audiosrc.link(self._queue)

        self._available = self._audiosrc is not None
        self.notify("available")

    @GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READABLE)
    def available(self) -> bool:
        return self._available

    def _set_property_available(self, value: bool) -> None:
        if value == self._available:
            return
        self._available = value
        self.notify("available")

    @property
    def recording_in_progress(self) -> bool:
        _, pipeline_state, _ = self._pipeline.get_state(timeout=100)
        return pipeline_state != Gst.State.NULL

    @property
    def audio_file_is_valid(self) -> bool:
        return self._output_file_valid

    @property
    def audio_file_abspath(self) -> Path:
        return self._file_path

    @property
    def audio_file_uri(self) -> str:
        uri = self._file_path.as_uri()
        assert uri is not None
        return uri

    def cleanup(self) -> None:
        del self._error_callback
        self._file_path.unlink(missing_ok=True)
        self._pipeline.get_bus().disconnect(self._id)

    def start_recording(self) -> None:
        assert self._filesink is not None

        if self._new_recording:
            self._new_recording = False
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            self._file_name = f"voice-message-{timestamp}.{self._filetype}"
            self._file_path = configpaths.get_temp_dir() / self._file_name

        if not self._file_path.parent.exists():
            self._handle_error_output_dir_inaccessible()
            return

        self._output_file_counter += 1
        self._filesink.set_property(
            "location", f"{self._file_path}.part{self._output_file_counter}"
        )
        log.debug(
            "Creating new recording file: %s", self._filesink.get_property("location")
        )
        self._pipeline.set_state(Gst.State.PLAYING)
        message = self._pipeline.get_bus().pop_filtered(Gst.MessageType.ERROR)
        if message is not None:
            self._handle_error_on_start(message)
        self._clock = self._pipeline.get_pipeline_clock()
        self._rec_time["start"] = self._clock.get_time()

    def stop_recording(self) -> None:
        if not self.recording_in_progress:
            return

        assert self._clock is not None

        self._rec_time["total"] += self._clock.get_time() - self._rec_time["start"]

        self._pipeline.send_event(Gst.Event.new_eos())
        message = self._pipeline.get_bus().timed_pop_filtered(
            Gst.CLOCK_TIME_NONE, Gst.MessageType.EOS | Gst.MessageType.ERROR
        )
        self._pipeline.set_state(Gst.State.NULL)
        assert message is not None
        self._handle_error_on_stop(message)

        self._merge_output_files()

    def stop_and_reset(self) -> None:
        if self.recording_in_progress:
            self.stop_recording()

        self._new_recording = True
        self._output_file_valid = False
        self._output_file_counter = 0
        self._output_files_invalid = []
        self._samples = deque([0] * self._num_samples, maxlen=self._num_samples)
        self._rec_time = {"start": 0, "total": 0}

    def recording_time(self) -> int:
        delta = 0
        if self.recording_in_progress and self._clock is not None:
            delta = self._clock.get_time() - self._rec_time["start"]
        return self._rec_time["total"] + delta

    def request_new_sample(self) -> None:
        if self._buffer_level != 0:
            samples_avg = math.fsum(list(self._samples_buffer)) / self._buffer_level
            self._buffer_level = 0
        else:
            samples_avg = 0
        self._samples_buffer = deque(
            [0.0] * self._num_samples_buffer, maxlen=self._num_samples_buffer
        )
        self._samples.appendleft(samples_avg)

    def recording_samples(self) -> list[tuple[float, float]]:
        sample_list = reversed(list(self._samples))
        result: list[tuple[float, float]] = []
        for sample in sample_list:
            result.append((sample, sample))
        return result

    def _is_error(self, message: Gst.Message | None) -> bool:
        if message is not None and message.type == Gst.MessageType.ERROR:
            self._log_error(message)
            return True
        return False

    def _log_error(self, message: Gst.Message | None) -> None:
        if message is None:
            return

        structure = message.get_structure()
        if structure is not None:
            gerror = structure.get_value("gerror")
            debug = structure.get_value("debug")

            assert gerror is not None
            log.error("gerror code: %s", gerror.code)
            log.error("gerror domain: %s", gerror.domain)
            log.error("debug: %s", debug)

    def _error_info(self, message: Gst.Message | None) -> tuple[str | None, int | None]:
        if message is None:
            return None, None
        structure = message.get_structure()
        assert structure is not None
        gerror = structure.get_value("gerror")
        assert gerror is not None
        return gerror.domain, gerror.code

    def _handle_error_on_start(self, message: Gst.Message) -> None:
        assert message is not None
        if not self._is_error(message):
            return

        log.error("Error on starting the recording.")
        self._output_files_invalid.append(self._output_file_counter)

        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 5:
                # GST_RESOURCE_ERROR_OPEN_READ (5)
                # used when resource fails to open for reading.
                self._error_callback(
                    GST_ERROR_ON_START, _("Is a microphone plugged in and accessible?")
                )

    def _handle_error_on_recording(self, message: Gst.Message | None):
        if message is not None and not self._is_error(message):
            return

        log.error("Error during the recording.")

        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 9:
                # GST_RESOURCE_ERROR_READ (9)
                # used when the resource can't be read from.
                self._error_callback(
                    GST_ERROR_ON_RECORDING,
                    _("Is a microphone plugged in and accessible?"),
                )

        if domain == "gst-stream-error-quark":
            if code == 1:
                # GST_STREAM_ERROR_FAILED (1)
                # a general error which doesn't fit in any other category.
                # Make sure you add a custom message to the error call.
                self._error_callback(
                    GST_ERROR_ON_RECORDING, _("Error in audio data stream.")
                )

    def _handle_error_on_stop(self, message: Gst.Message | None) -> None:
        if message is not None and not self._is_error(message):
            return
        # TODO
        log.error("Error when stopping the recording.")

    def _handle_error_on_merging(self, message: Gst.Message | None) -> None:
        if message is not None and not self._is_error(message):
            return

        # TODO
        log.error("Error when merging the recordings.")

    def _handle_error_output_dir_inaccessible(self) -> None:
        log.error(
            "Voice message lost. Temporary output folder %s not accessible",
            self._file_path.parent,
        )
        error_message = _("Voice message could not be saved. Please try again.")
        self._error_callback(GST_ERROR_ON_MERGING, error_message)
        self.stop_and_reset()

    def _on_gst_message(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        structure = message.get_structure()
        if structure is None:
            return
        message_string = structure.to_string()

        if message.type == Gst.MessageType.ELEMENT:
            name = structure.get_name()
            if name == "level":
                if not structure.has_field("rms"):
                    return

                rms_values = structure.get_value("rms")
                assert rms_values is not None
                if len(rms_values) > 0:
                    rms_value = rms_values[0]
                    self._rms = math.pow(10, rms_value / 10 / 2)
                    self._samples_buffer.appendleft(self._rms)
                    self._buffer_level += 1 % self._num_samples
            log.debug("gst element message: %s", message_string)

        elif message.type == Gst.MessageType.ERROR:
            self._handle_error_on_recording(message)

        elif message.type == Gst.MessageType.EOS:
            pass

    def _switch_sources(self) -> None:
        assert self._queue is not None
        assert not self.recording_in_progress

        if self._audiosrc is not None:
            self._audiosrc.unlink(self._queue)
            self._pipeline.remove(self._audiosrc)

        self._audiosrc = self._create_new_audio_src()
        if self._audiosrc is not None:
            self._pipeline.add(self._audiosrc)
            self._audiosrc.link(self._queue)

        self._available = self._audiosrc is not None
        self.notify("available")

    def _create_new_audio_src(self) -> Gst.Element | None:
        device = self._extract_first_word(app.settings.get("audio_input_device"))

        if sys.platform == "win32" and device == "autoaudiosrc":
            device = "wasapisrc"

        if device == "audiotestsrc":
            log.error('Audio device "audiotestsrc" not supported')
            return None

        audiosrc = Gst.ElementFactory.make(device)
        if audiosrc is None:
            log.debug("Unable to create audio source for: %s", device)
        else:
            log.debug("Created audio source for: %s", device)
            if device == "wasapisrc":
                audiosrc.set_property("role", "comms")

        return audiosrc

    def _on_audio_input_device_changed(self, device: str, *args: Any) -> None:
        log.debug("Switching to input device: %s", device)
        self._switch_sources()

    def _file_merge_required(self) -> bool:
        return self._output_file_counter > 1

    def _merge_opus_m4a_command(self) -> str:
        log.info("Merging opus files started")

        # Use as_posix() on file path to convert "\" to "/" on Windows
        sources = ""
        for i in range(1, self._output_file_counter + 1):
            if i in self._output_files_invalid:
                continue
            source = " ! ".join(
                [
                    f"filesrc location={self._file_path.as_posix()}.part{i}",
                    "qtdemux",
                    "opusdec",
                    "c. ",
                ]
            )
            sources += source

        command = " ! ".join(
            [
                "concat name=c",
                "queue",
                "audioconvert",
                "audioresample",
                "opusenc audio-type=voice",
                "mp4mux ",
                f"filesink location={self._file_path.as_posix()} {sources}",
            ]
        )
        return command

    def _merge_output_files(self) -> None:
        if not self._file_path.parent.exists():
            self._handle_error_output_dir_inaccessible()
            return

        command = self._merge_opus_m4a_command()
        pipeline = Gst.parse_launch(command)

        # Remove the original output file first before writing into it
        if self._file_path.exists():
            self._file_path.unlink()

        pipeline.set_state(Gst.State.PLAYING)

        bus = pipeline.get_bus()
        if bus is None:
            return

        message = bus.timed_pop_filtered(
            Gst.CLOCK_TIME_NONE, Gst.MessageType.EOS | Gst.MessageType.ERROR
        )
        pipeline.set_state(Gst.State.NULL)

        self._handle_error_on_merging(message)
        self._output_file_valid = True
        log.debug("Merging files finished")

    @staticmethod
    def _extract_first_word(gst_cmd: str) -> str:
        """
        Gajim gives us a string with an audio src plus additional parameters
        such as volume. This method takes only the first word == audio src.
        """
        if " " in gst_cmd:
            idx = gst_cmd.index(" ")
            gst_cmd = gst_cmd[:idx]
        return gst_cmd



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime

from nbxmpp import JID
from sqlalchemy import types
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.types import TypeEngine

from gajim.common.storage.base import EpochTimestampType
from gajim.common.storage.base import JIDType


class Base(DeclarativeBase):
    type_annotation_map: dict[type[Any], TypeEngine[Any]] = {
        float: types.REAL(),
        str: types.TEXT(),
        int: types.INTEGER(),
    }


class Event(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = "event"

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    account: Mapped[str] = mapped_column()
    jid: Mapped[JID] = mapped_column(JIDType)
    event: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)
    data: Mapped[str] = mapped_column()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import logging

from gi.repository import Adw
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.modules.muc.util import MucInfoResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.const import MUC_DISCO_ERRORS
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _
from gajim.common.util.muc import get_group_chat_nick
from gajim.common.util.standards import get_rfc5646_lang

from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.groupchat_nick_chooser import GroupChatNickChooser
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.groupchat_join")


class GroupchatJoin(GajimAppWindow):
    def __init__(self, account: str, jid: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="GroupchatJoin",
            title=_("Join Group Chat"),
            default_width=500,
            default_height=550,
            add_window_padding=True,
            header_bar=True,
        )

        self._destroyed = False
        self.account = account
        self.jid = jid
        self._redirected = False

        self._main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=18)
        self._main_box.set_valign(Gtk.Align.FILL)

        self._muc_info_box = GroupChatInfoScrolled(account)

        self._stack = Gtk.Stack()
        self._stack.add_named(self._muc_info_box, "info")
        self._stack.add_named(ProgressPage(), "progress")
        self._stack.add_named(ErrorPage(), "error")

        self._stack.set_visible_child_name("progress")

        self._connect(self._stack, "notify::visible-child-name", self._on_page_changed)
        self._main_box.append(self._stack)

        self._nick_chooser = GroupChatNickChooser()

        self._join_button = Gtk.Button.new_with_mnemonic(_("_Join"))
        self._join_button.set_halign(Gtk.Align.END)
        self._join_button.set_sensitive(False)
        self._join_button.add_css_class("suggested-action")
        self._connect(self._join_button, "clicked", self._on_join)

        join_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        join_box.add_css_class("linked")
        join_box.set_halign(Gtk.Align.END)
        join_box.append(self._nick_chooser)
        join_box.append(self._join_button)

        self._main_box.append(join_box)

        self.set_child(self._main_box)

        client = app.get_client(self.account)
        client.get_module("Discovery").disco_muc(
            jid,
            allow_redirect=True,
            request_vcard=True,
            callback=self._disco_info_received,  # type: ignore
        )

    def _on_page_changed(self, stack: Gtk.Stack, _param: Any) -> None:
        name = stack.get_visible_child_name()
        self._join_button.set_sensitive(name == "info")
        self._nick_chooser.set_sensitive(name == "info")

    @ensure_not_destroyed
    def _disco_info_received(self, task: Task) -> None:
        try:
            result = cast(MucInfoResult, task.finish())
        except StanzaError as error:
            log.info("Disco %s failed: %s", error.jid, error.get_text())
            self._set_error(error)
            return

        if result.redirected:
            self.jid = str(result.info.jid)

        if result.info.is_muc:
            self._muc_info_box.set_from_disco_info(result.info)
            assert result.info.jid is not None
            nickname = get_group_chat_nick(self.account, result.info.jid)
            self._nick_chooser.set_text(nickname)
            self.set_default_widget(self._join_button)
            self._stack.set_visible_child_name("info")

        else:
            self._set_error_from_code("not-muc-service")

    def _show_error_page(self, text: str) -> None:
        error_page = cast(ErrorPage, self._stack.get_child_by_name("error"))
        error_page.set_text(text)
        self._stack.set_visible_child_name("error")

    def _set_error(self, error: StanzaError) -> None:
        text = MUC_DISCO_ERRORS.get(error.condition or "", to_user_string(error))
        if error.condition == "gone":
            reason = error.get_text(get_rfc5646_lang())
            if reason:
                text = f"{text}:\n{reason}"
        self._show_error_page(text)

    def _set_error_from_code(self, error_code: str) -> None:
        self._show_error_page(MUC_DISCO_ERRORS[error_code])

    def _on_join(self, _button: Gtk.Button) -> None:
        nickname = self._nick_chooser.get_text()

        app.window.show_add_join_groupchat(self.account, self.jid, nickname=nickname)
        self.close()

    def _cleanup(self) -> None:
        self._destroyed = True


class ErrorPage(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=18)
        self.set_vexpand(True)
        self.set_homogeneous(True)
        error_icon = Gtk.Image.new_from_icon_name("lucide-circle-x-symbolic")
        error_icon.set_valign(Gtk.Align.END)

        self._error_label = Gtk.Label(
            wrap=True,
            width_request=150,
            justify=Gtk.Justification.CENTER,
            valign=Gtk.Align.START,
        )
        self._error_label.add_css_class("title-3")

        self.append(error_icon)
        self.append(self._error_label)

    def set_text(self, text: str) -> None:
        self._error_label.set_text(text)


class ProgressPage(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(
            self,
            orientation=Gtk.Orientation.VERTICAL,
            spacing=18,
            vexpand=True,
            homogeneous=True,
        )
        spinner = Adw.Spinner(halign=Gtk.Align.CENTER)
        self.append(spinner)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.modules.vcard4 import EmailProperty
from nbxmpp.modules.vcard4 import OrgProperty
from nbxmpp.modules.vcard4 import RoleProperty
from nbxmpp.modules.vcard4 import TelProperty
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import VCard

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.util.status import get_uf_show
from gajim.common.util.user_strings import get_time_zone_string

from gajim.gtk.avatar import get_show_circle
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import convert_surface_to_texture
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.misc import open_uri


@Gtk.Template(string=get_ui_string("contact_popover.ui"))
class ContactPopover(Gtk.Popover, SignalManager):
    __gtype_name__ = "ContactPopover"

    _avatar: Gtk.Image = Gtk.Template.Child()
    _name: Gtk.Label = Gtk.Template.Child()
    _role: Gtk.Label = Gtk.Template.Child()
    _org: Gtk.Label = Gtk.Template.Child()
    _info_listbox: Gtk.ListBox = Gtk.Template.Child()
    _subscription: ContactPopoverInfoRow = Gtk.Template.Child()
    _status: ContactPopoverInfoRow = Gtk.Template.Child()
    _status_message: ContactPopoverInfoRow = Gtk.Template.Child()
    _xmpp_address: ContactPopoverInfoRow = Gtk.Template.Child()
    _email: ContactPopoverInfoRow = Gtk.Template.Child()
    _tel: ContactPopoverInfoRow = Gtk.Template.Child()
    _timezone: ContactPopoverInfoRow = Gtk.Template.Child()
    _contact_details_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self, contact: BareContact) -> None:
        Gtk.Popover.__init__(self)
        SignalManager.__init__(self)

        self._contact = contact

        scale = self.get_scale_factor()

        texture = contact.get_avatar(AvatarSize.TOOLTIP, scale)
        self._avatar.set_pixel_size(AvatarSize.TOOLTIP)
        self._avatar.set_from_paintable(texture)

        if contact.is_self:
            self._name.set_label(_("Note to myself"))
        else:
            self._name.set_label(contact.name)

        if not contact.is_self:
            if contact.subscription in ("none", "to"):
                self._subscription.set_label(
                    _("You don't share your status with this contact")
                )

            if contact.subscription == ("from"):
                self._subscription.set_label(
                    _("This contact does not share their status with you")
                )

        # Status
        surface = get_show_circle(
            self._contact.show, AvatarSize.SHOW_CIRCLE, self.get_scale_factor()
        )
        icon = convert_surface_to_texture(surface)
        status_text = get_uf_show(self._contact.show.value)
        if idle_time_text := self._get_idle_time():
            status_text += f" ({idle_time_text})"

        self._status.set_label(status_text)
        self._status.set_icon(icon)

        if self._contact.status:
            self._status_message.set_label(self._contact.status)

        self._xmpp_address.set_label(str(self._contact.jid), link_scheme="xmpp")

        self._connect(
            self._contact_details_button, "clicked", self._on_contact_details_clicked
        )

        client = app.get_client(contact.account)
        vcard = client.get_module("VCard4").request_vcard(
            jid=self._contact.jid, callback=self._on_vcard_received, use_cache=True
        )
        if vcard is not None:
            self._set_vcard_rows(vcard)

        app.plugin_manager.extension_point("contact_tooltip_populate", self, contact)

    def _on_vcard_received(self, jid: JID, vcard: VCard) -> None:
        self._set_vcard_rows(vcard)

    def _set_vcard_rows(self, vcard: VCard) -> None:
        for prop in vcard.get_properties():
            match prop:
                case RoleProperty():
                    self._role.set_label(prop.value)
                    self._role.set_visible(True)
                case OrgProperty():
                    self._org.set_label(prop.values[0])
                    self._org.set_visible(True)
                case EmailProperty():
                    self._email.set_label(prop.value, link_scheme="mailto")
                case TelProperty():
                    self._tel.set_label(prop.value, link_scheme="tel")
                case TzProperty():
                    self._timezone.set_label(get_time_zone_string(prop))
                case _:
                    pass

    def _get_idle_time(self) -> str | None:
        if self._contact.idle_datetime is None:
            return None

        current = dt.datetime.now()
        if self._contact.idle_datetime.date() == current.date():
            format_string = app.settings.get("time_format")
            formatted = self._contact.idle_datetime.strftime(format_string)
        else:
            format_string = app.settings.get("date_time_format")
            formatted = self._contact.idle_datetime.strftime(format_string)
        return _("last seen: %s") % formatted

    def _on_contact_details_clicked(self, _button: Gtk.Button) -> None:
        self.popdown()
        account_jid_params = AccountJidParam(
            account=self._contact.account, jid=self._contact.jid
        )
        app.window.activate_action(
            "win.chat-contact-info", account_jid_params.to_variant()
        )

    def do_unroot(self) -> None:
        self._disconnect_all()
        app.check_finalize(self)
        Gtk.Popover.do_unroot(self)


@Gtk.Template(string=get_ui_string("contact_popover_info_row.ui"))
class ContactPopoverInfoRow(Gtk.ListBoxRow):
    __gtype_name__ = "ContactPopoverInfoRow"

    _image: Gtk.Image = Gtk.Template.Child()
    _label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.ListBoxRow.__init__(self)

        self._icon: Gdk.Paintable | None = None
        self._icon_name: str | None = None
        self._text: str = ""

    @GObject.Property(type=Gdk.Paintable)
    def icon(self) -> Gdk.Paintable | None:  # pyright: ignore
        return self._icon

    @icon.setter
    def icon(self, icon: Gdk.Paintable | None) -> None:
        self.set_icon(icon)

    @GObject.Property(type=str)
    def icon_name(self) -> str | None:  # pyright: ignore
        return self._icon_name

    @icon_name.setter
    def icon_name(self, icon_name: str | None) -> None:
        self.set_icon(icon_name)

    @GObject.Property(type=str)
    def text(self) -> str:  # pyright: ignore
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        self.set_label(text)

    def set_icon(self, icon: str | Gdk.Paintable | None) -> None:
        self._icon_name = None
        self._icon = None
        if icon is None:
            self._image.set_from_icon_name(None)
            self._image.set_from_paintable(None)
            return

        if isinstance(icon, str):
            self._image.set_from_icon_name(icon)
            self._icon_name = icon
        else:
            self._image.set_from_paintable(icon)
            self._icon = icon

    def set_label(self, text: str, link_scheme: str | None = None) -> None:
        self._text = text
        self.set_visible(bool(text))

        if link_scheme:
            self._label.set_markup(f"<a href='{link_scheme}:{text}'>{text}</a>")
        else:
            self._label.set_text(text)

        self._label.set_selectable(bool(link_scheme))

        if len(text) > 30:
            self._label.set_tooltip_text(text)

    @Gtk.Template.Callback()
    @staticmethod
    def _on_activate_link(label: Gtk.Label, *args: Any) -> int:
        open_uri(label.get_current_uri() or label.get_text())
        return Gdk.EVENT_STOP

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging

from gi.repository import Gio
from gi.repository import Gtk

from gajim.common.i18n import _

from gajim.gtk.preference.certificate import CertificatePage
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.certificate_dialog")


class CertificateDialog(GajimAppWindow):
    def __init__(
        self, transient_for: Gtk.Window | None, account: str, cert: Gio.TlsCertificate
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="CertificateDialog",
            title=_("Certificate"),
            default_width=600,
            default_height=800,
            transient_for=transient_for,
            add_window_padding=True,
            header_bar=True,
        )

        self.account = account

        self.set_child(CertificatePage(account, cert))

    def _cleanup(self) -> None:
        pass



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
# Copyright (C) 2015 Tarek Galal <tare2.galal@gmail.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import NamedTuple

import sqlite3
import time
from collections import namedtuple
from pathlib import Path

from omemo_dr.const import OMEMOTrust
from omemo_dr.ecc.djbec import CurvePublicKey
from omemo_dr.ecc.djbec import DjbECPrivateKey
from omemo_dr.exceptions import InvalidKeyIdException
from omemo_dr.identitykey import IdentityKey
from omemo_dr.identitykeypair import IdentityKeyPair
from omemo_dr.state.prekeyrecord import PreKeyRecord
from omemo_dr.state.sessionrecord import SessionRecord
from omemo_dr.state.signedprekeyrecord import SignedPreKeyRecord
from omemo_dr.state.store import Store
from omemo_dr.structs import IdentityInfo

from gajim.common import app
from gajim.common.modules.util import LogAdapter


def _convert_identity_key(key: bytes) -> IdentityKey | None:
    if not key:
        return
    return IdentityKey(CurvePublicKey(key[1:]))


def _convert_record(record: bytes) -> SessionRecord:
    return SessionRecord(serialized=record)


sqlite3.register_converter("pk", _convert_identity_key)
sqlite3.register_converter("session_record", _convert_record)


class OMEMOStorage(Store):
    def __init__(self, account: str, db_path: Path, log: LogAdapter) -> None:
        self._log = log
        self._account = account
        self._con = sqlite3.connect(db_path, detect_types=sqlite3.PARSE_COLNAMES)
        self._con.row_factory = self._namedtuple_factory
        self.create_db()
        self.migrate_db()

        self._con.execute("PRAGMA secure_delete=1")
        self._con.execute("PRAGMA synchronous=NORMAL;")
        mode = self._con.execute("PRAGMA journal_mode;").fetchone()[0]

        # WAL is a persistent DB mode, don't override it if user has set it
        if mode != "wal":
            self._con.execute("PRAGMA journal_mode=MEMORY;")
        self._con.commit()

    def _is_blind_trust_enabled(self) -> bool:
        return app.settings.get_account_setting(self._account, "omemo_blind_trust")

    @staticmethod
    def _namedtuple_factory(cursor: sqlite3.Cursor, row: tuple[Any, ...]) -> NamedTuple:
        fields: list[str] = []
        for col in cursor.description:
            if col[0] == "_id":
                fields.append("id")
            elif "strftime" in col[0]:
                fields.append("formated_time")
            elif "MAX" in col[0] or "COUNT" in col[0]:
                col_name = col[0].replace("(", "_")
                col_name = col_name.replace(")", "")
                fields.append(col_name.lower())
            else:
                fields.append(col[0])
        return namedtuple("Row", fields)(*row)  # pyright: ignore

    def user_version(self) -> int:
        return self._con.execute("PRAGMA user_version").fetchone()[0]

    def create_db(self) -> None:
        if self.user_version() == 0:
            create_tables = """
                CREATE TABLE IF NOT EXISTS secret (
                    device_id INTEGER, public_key BLOB, private_key BLOB);

                CREATE TABLE IF NOT EXISTS identities (
                    _id INTEGER PRIMARY KEY AUTOINCREMENT, recipient_id TEXT,
                    registration_id INTEGER, public_key BLOB,
                    timestamp INTEGER, trust INTEGER,
                    shown INTEGER DEFAULT 0);

                CREATE UNIQUE INDEX IF NOT EXISTS
                    public_key_index ON identities (public_key, recipient_id);

                CREATE TABLE IF NOT EXISTS prekeys(
                    _id INTEGER PRIMARY KEY AUTOINCREMENT,
                    prekey_id INTEGER UNIQUE, sent_to_server BOOLEAN,
                    record BLOB);

                CREATE TABLE IF NOT EXISTS signed_prekeys (
                    _id INTEGER PRIMARY KEY AUTOINCREMENT,
                    prekey_id INTEGER UNIQUE,
                    timestamp NUMERIC DEFAULT CURRENT_TIMESTAMP, record BLOB);

                CREATE TABLE IF NOT EXISTS sessions (
                    _id INTEGER PRIMARY KEY AUTOINCREMENT,
                    recipient_id TEXT, device_id INTEGER,
                    record BLOB, timestamp INTEGER, active INTEGER DEFAULT 1,
                    UNIQUE(recipient_id, device_id));

                """

            create_db_sql = """
                BEGIN TRANSACTION;
                %s
                PRAGMA user_version=12;
                END TRANSACTION;
                """ % (create_tables)
            self._con.executescript(create_db_sql)

    def migrate_db(self) -> None:
        """Migrates the DB"""

        # Find all double entries and delete them
        if self.user_version() < 2:
            delete_dupes = """ DELETE FROM identities WHERE _id not in (
                                SELECT MIN(_id)
                                FROM identities
                                GROUP BY
                                recipient_id, public_key
                                );
                            """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=2;
                    END TRANSACTION;
                """
                % (delete_dupes)
            )

        if self.user_version() < 3:
            # Create a UNIQUE INDEX so every public key/recipient_id tuple
            # can only be once in the db
            add_index = """ CREATE UNIQUE INDEX IF NOT EXISTS
                            public_key_index
                            ON identities (public_key, recipient_id);
                        """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=3;
                    END TRANSACTION;
                """
                % (add_index)
            )

        if self.user_version() < 4:
            # Adds column 'active' to the sessions table
            add_active = """ ALTER TABLE sessions
                             ADD COLUMN active INTEGER DEFAULT 1;
                         """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=4;
                    END TRANSACTION;
                """
                % (add_active)
            )

        if self.user_version() < 5:
            # Adds DEFAULT Timestamp
            add_timestamp = """
                DROP TABLE signed_prekeys;
                CREATE TABLE IF NOT EXISTS signed_prekeys (
                    _id INTEGER PRIMARY KEY AUTOINCREMENT,
                    prekey_id INTEGER UNIQUE,
                    timestamp NUMERIC DEFAULT CURRENT_TIMESTAMP, record BLOB);
                ALTER TABLE identities ADD COLUMN shown INTEGER DEFAULT 0;
                UPDATE identities SET shown = 1;
            """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=5;
                    END TRANSACTION;
                """
                % (add_timestamp)
            )

        if self.user_version() < 6:
            # Move secret data into own table
            # We add +1 to registration id because we did that in other code in
            # earlier versions. On this migration we correct this mistake now.
            move = """
                CREATE TABLE IF NOT EXISTS secret (
                    device_id INTEGER, public_key BLOB, private_key BLOB);
                INSERT INTO secret (device_id, public_key, private_key)
                SELECT registration_id + 1, public_key, private_key
                FROM identities
                WHERE recipient_id = -1;
            """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=6;
                    END TRANSACTION;
                """
                % move
            )

        if self.user_version() < 7:
            # Convert old device ids to integer
            convert = """
                UPDATE secret SET device_id = device_id % 2147483646;
            """

            self._con.executescript(
                """ BEGIN TRANSACTION;
                    %s
                    PRAGMA user_version=7;
                    END TRANSACTION;
                """
                % convert
            )

        if self.user_version() < 8:
            # Sanitize invalid BLOBs from the python2 days
            query_keys = """SELECT recipient_id,
                            registration_id,
                            CAST(public_key as BLOB) as public_key,
                            CAST(private_key as BLOB) as private_key,
                            timestamp, trust, shown
                            FROM identities"""
            rows = self._con.execute(query_keys).fetchall()

            delete = "DELETE FROM identities"
            self._con.execute(delete)

            insert = """INSERT INTO identities (
                        recipient_id, registration_id, public_key, private_key,
                        timestamp, trust, shown)
                        VALUES (?, ?, ?, ?, ?, ?, ?)"""
            for row in rows:
                try:
                    self._con.execute(insert, row)
                except Exception as error:
                    self._log.warning(error)
            self._con.execute("PRAGMA user_version=8")
            self._con.commit()

        if self.user_version() < 9:
            # Sanitize invalid BLOBs from the python2 days
            query_keys = """SELECT device_id,
                            CAST(public_key as BLOB) as public_key,
                            CAST(private_key as BLOB) as private_key
                            FROM secret"""
            rows = self._con.execute(query_keys).fetchall()

            delete = "DELETE FROM secret"
            self._con.execute(delete)

            insert = """INSERT INTO secret (device_id, public_key, private_key)
                        VALUES (?, ?, ?)"""
            for row in rows:
                try:
                    self._con.execute(insert, row)
                except Exception as error:
                    self._log.warning(error)
            self._con.execute("PRAGMA user_version=9")
            self._con.commit()

        if self.user_version() < 10:
            # Sanitize invalid BLOBs from the python2 days
            query_keys = """SELECT _id,
                            recipient_id,
                            device_id,
                            CAST(record as BLOB) as record,
                            timestamp,
                            active
                            FROM sessions"""
            rows = self._con.execute(query_keys).fetchall()

            delete = "DELETE FROM sessions"
            self._con.execute(delete)

            insert = """INSERT INTO sessions (_id, recipient_id, device_id,
                                              record, timestamp, active)
                        VALUES (?, ?, ?, ?, ?, ?)"""
            for row in rows:
                try:
                    self._con.execute(insert, row)
                except Exception as error:
                    self._log.warning(error)
            self._con.execute("PRAGMA user_version=10")
            self._con.commit()

        if self.user_version() < 11:
            # Sanitize invalid BLOBs from the python2 days
            query_keys = """SELECT _id,
                            prekey_id,
                            sent_to_server,
                            CAST(record as BLOB) as record
                            FROM prekeys"""
            rows = self._con.execute(query_keys).fetchall()

            delete = "DELETE FROM prekeys"
            self._con.execute(delete)

            insert = """INSERT INTO prekeys (
                        _id, prekey_id, sent_to_server, record)
                        VALUES (?, ?, ?, ?)"""
            for row in rows:
                try:
                    self._con.execute(insert, row)
                except Exception as error:
                    self._log.warning(error)
            self._con.execute("PRAGMA user_version=11")
            self._con.commit()

        if self.user_version() < 12:
            # Sanitize invalid BLOBs from the python2 days
            query_keys = """SELECT _id,
                            prekey_id,
                            timestamp,
                            CAST(record as BLOB) as record
                            FROM signed_prekeys"""
            rows = self._con.execute(query_keys).fetchall()

            delete = "DELETE FROM signed_prekeys"
            self._con.execute(delete)

            insert = """INSERT INTO signed_prekeys (
                        _id, prekey_id, timestamp, record)
                        VALUES (?, ?, ?, ?)"""
            for row in rows:
                try:
                    self._con.execute(insert, row)
                except Exception as error:
                    self._log.warning(error)
            self._con.execute("PRAGMA user_version=12")
            self._con.commit()

    def load_signed_pre_key(self, signed_pre_key_id: int) -> SignedPreKeyRecord:
        query = "SELECT record FROM signed_prekeys WHERE prekey_id = ?"
        result = self._con.execute(query, (signed_pre_key_id,)).fetchone()
        if result is None:
            raise InvalidKeyIdException(
                "No such signedprekeyrecord! %s " % signed_pre_key_id
            )
        return SignedPreKeyRecord.from_bytes(result.record)

    def load_signed_pre_keys(self) -> list[SignedPreKeyRecord]:
        query = "SELECT record FROM signed_prekeys"
        results = self._con.execute(query).fetchall()
        return [SignedPreKeyRecord.from_bytes(row.record) for row in results]

    def store_signed_pre_key(
        self, signed_pre_key_id: int, signed_pre_key_record: SignedPreKeyRecord
    ) -> None:
        query = "INSERT INTO signed_prekeys (prekey_id, record) VALUES(?,?)"
        self._con.execute(query, (signed_pre_key_id, signed_pre_key_record.serialize()))
        self._con.commit()

    def contains_signed_pre_key(self, signed_pre_key_id: int) -> bool:
        query = "SELECT record FROM signed_prekeys WHERE prekey_id = ?"
        result = self._con.execute(query, (signed_pre_key_id,)).fetchone()
        return result is not None

    def remove_signed_pre_key(self, signed_pre_key_id: int) -> None:
        query = "DELETE FROM signed_prekeys WHERE prekey_id = ?"
        self._con.execute(query, (signed_pre_key_id,))
        self._con.commit()

    def get_current_signed_pre_key_id(self) -> int:
        query = "SELECT MAX(prekey_id) FROM signed_prekeys"
        result = self._con.execute(query).fetchone()
        assert result is not None
        return result.max_prekey_id

    def get_signed_pre_key_timestamp(self, signed_pre_key_id: int) -> int:
        query = """SELECT strftime('%s', timestamp) FROM
                   signed_prekeys WHERE prekey_id = ?"""

        result = self._con.execute(query, (signed_pre_key_id,)).fetchone()
        if result is None:
            raise InvalidKeyIdException(
                "No such signedprekeyrecord! %s" % signed_pre_key_id
            )

        return result.formated_time

    def remove_old_signed_pre_keys(self, timestamp: int) -> None:
        query = """DELETE FROM signed_prekeys
                   WHERE prekey_id NOT IN (
                    SELECT prekey_id FROM signed_prekeys
                    ORDER BY timestamp desc LIMIT 3
                   )"""

        self._con.execute(query)
        self._con.commit()

    def load_session(self, recipient_id: str, device_id: int) -> SessionRecord:
        query = """SELECT record as "record [session_record]"
                   FROM sessions WHERE recipient_id = ? AND device_id = ?"""
        result = self._con.execute(query, (recipient_id, device_id)).fetchone()
        return result.record if result is not None else SessionRecord()

    def get_jid_from_device(self, device_id: int) -> str | None:
        query = """SELECT recipient_id
                   FROM sessions WHERE device_id = ?"""
        result = self._con.execute(query, (device_id,)).fetchone()
        return result.recipient_id if result is not None else None

    def get_active_device_tuples(self):
        query = """SELECT recipient_id, device_id
                   FROM sessions WHERE active = 1"""
        return self._con.execute(query).fetchall()

    def store_session(
        self, recipient_id: str, device_id: int, session_record: SessionRecord
    ) -> None:
        query = """INSERT INTO sessions(recipient_id, device_id, record)
                   VALUES(?,?,?)"""
        try:
            self._con.execute(
                query, (recipient_id, device_id, session_record.serialize())
            )
        except sqlite3.IntegrityError:
            query = """UPDATE sessions SET record = ?
                       WHERE recipient_id = ? AND device_id = ?"""
            self._con.execute(
                query, (session_record.serialize(), recipient_id, device_id)
            )

        self._con.commit()

    def contains_session(self, recipient_id: str, device_id: int) -> bool:
        query = """SELECT record FROM sessions
                   WHERE recipient_id = ? AND device_id = ?"""
        result = self._con.execute(query, (recipient_id, device_id)).fetchone()
        return result is not None

    def delete_session(self, recipient_id: str, device_id: int) -> None:
        self._log.info("Delete session for %s %s", recipient_id, device_id)
        query = "DELETE FROM sessions WHERE recipient_id = ? AND device_id = ?"
        self._con.execute(query, (recipient_id, device_id))
        self._con.commit()

    def delete_all_sessions(self, recipient_id: str) -> None:
        query = "DELETE FROM sessions WHERE recipient_id = ?"
        self._con.execute(query, (recipient_id,))
        self._con.commit()

    def get_identity_infos(self, recipient_ids: str | list[str]) -> list[IdentityInfo]:
        if isinstance(recipient_ids, str):
            recipient_ids = [recipient_ids]

        query = """SELECT recipient_id,
                          public_key as "public_key [pk]",
                          trust,
                          timestamp
                   FROM identities
                   WHERE recipient_id IN ({})""".format(
            ", ".join(["?"] * len(recipient_ids))
        )
        i_results = self._con.execute(query, recipient_ids).fetchall()

        query = """SELECT device_id,
                          record as "record [session_record]",
                          active
                   FROM sessions WHERE recipient_id IN ({})""".format(
            ", ".join(["?"] * len(recipient_ids))
        )
        s_results = self._con.execute(query, recipient_ids).fetchall()

        sessions: dict[IdentityKey, Any] = {}
        for s_result in s_results:
            if s_result.record.is_fresh():
                continue
            ik = s_result.record.get_session_state().get_remote_identity_key()
            sessions[ik] = s_result

        identity_infos: list[IdentityInfo] = []
        for i_result in i_results:
            session = sessions.get(i_result.public_key)
            if session is None:
                continue

            info = IdentityInfo(
                active=session.active,
                address=i_result.recipient_id,
                device_id=session.device_id,
                public_key=i_result.public_key,
                label="",
                last_seen=i_result.timestamp,
                trust=OMEMOTrust(i_result.trust),
            )
            identity_infos.append(info)

        return identity_infos

    def set_active_state(self, address: str, devicelist: list[int]) -> None:
        query = """
        UPDATE sessions SET active = 1
        WHERE recipient_id = ? AND device_id IN ({})""".format(
            ", ".join(["?"] * len(devicelist))
        )
        self._con.execute(query, (address,) + tuple(devicelist))

        query = """
        UPDATE sessions SET active = 0
        WHERE recipient_id = ? AND device_id NOT IN ({})""".format(
            ", ".join(["?"] * len(devicelist))
        )
        self._con.execute(query, (address,) + tuple(devicelist))
        self._con.commit()

    def set_inactive(self, address: str, device_id: int) -> None:
        query = """UPDATE sessions SET active = 0
                   WHERE recipient_id = ? AND device_id = ?"""
        self._con.execute(query, (address, device_id))
        self._con.commit()

    def get_inactive_sessions_keys(self, recipient_id: str) -> list[IdentityKey]:
        query = """SELECT record as "record [session_record]" FROM sessions
                   WHERE active = 0 AND recipient_id = ?"""
        results = self._con.execute(query, (recipient_id,)).fetchall()

        keys: list[IdentityKey] = []
        for result in results:
            key = result.record.get_session_state().get_remote_identity_key()
            keys.append(IdentityKey(key.get_public_key()))
        return keys

    def load_pre_key(self, pre_key_id: int) -> PreKeyRecord:
        query = """SELECT record FROM prekeys WHERE prekey_id = ?"""

        result = self._con.execute(query, (pre_key_id,)).fetchone()
        if result is None:
            raise Exception("No such prekeyRecord!")
        return PreKeyRecord.from_bytes(result.record)

    def load_pending_pre_keys(self) -> list[PreKeyRecord]:
        query = """SELECT record FROM prekeys"""
        result = self._con.execute(query).fetchall()
        return [PreKeyRecord.from_bytes(row.record) for row in result]

    def store_pre_key(self, pre_key_id: int, pre_key_record: PreKeyRecord) -> None:
        query = "INSERT INTO prekeys (prekey_id, record) VALUES(?,?)"
        self._con.execute(query, (pre_key_id, pre_key_record.serialize()))
        self._con.commit()

    def contains_pre_key(self, pre_key_id: int) -> bool:
        query = "SELECT record FROM prekeys WHERE prekey_id = ?"
        result = self._con.execute(query, (pre_key_id,)).fetchone()
        return result is not None

    def remove_pre_key(self, pre_key_id: int) -> None:
        query = "DELETE FROM prekeys WHERE prekey_id = ?"
        self._con.execute(query, (pre_key_id,))
        self._con.commit()

    def get_current_pre_key_id(self) -> int | None:
        query = "SELECT MAX(prekey_id) FROM prekeys"
        result = self._con.execute(query).fetchone()
        return result.max_prekey_id if result is not None else None

    def get_pre_key_count(self) -> int:
        query = "SELECT COUNT(prekey_id) FROM prekeys"
        return self._con.execute(query).fetchone().count_prekey_id

    def get_identity_key_pair(self) -> IdentityKeyPair:
        query = """SELECT public_key as "public_key [pk]", private_key
                   FROM secret LIMIT 1"""
        result = self._con.execute(query).fetchone()

        return IdentityKeyPair.new(
            result.public_key, DjbECPrivateKey(result.private_key)
        )

    def get_our_device_id(self) -> int:
        query = "SELECT device_id FROM secret LIMIT 1"
        result = self._con.execute(query).fetchone()
        assert result is not None
        return result.device_id

    def set_our_identity(
        self, device_id: int, identity_key_pair: IdentityKeyPair
    ) -> None:
        query = "SELECT * FROM secret"
        result = self._con.execute(query).fetchone()
        if result is not None:
            self._log.error("Trying to save secret key into non-empty secret table")
            return

        query = """INSERT INTO secret(device_id, public_key, private_key)
                   VALUES(?, ?, ?)"""

        public_key = identity_key_pair.get_public_key().get_public_key().serialize()
        private_key = identity_key_pair.get_private_key().serialize()
        self._con.execute(query, (device_id, public_key, private_key))
        self._con.commit()

    def save_identity(self, recipient_id: str, identity_key: IdentityKey) -> None:
        query = """INSERT INTO identities
                   (recipient_id, public_key, trust, shown)
                   VALUES(?, ?, ?, ?)"""
        if not self.contains_identity(recipient_id, identity_key):
            trust = self.get_default_trust(recipient_id)
            self._con.execute(
                query,
                (
                    recipient_id,
                    identity_key.get_public_key().serialize(),
                    trust,
                    1 if trust == OMEMOTrust.BLIND else 0,
                ),
            )
            self._con.commit()

    def contains_identity(self, recipient_id: str, identity_key: IdentityKey) -> bool:
        query = """SELECT * FROM identities WHERE recipient_id = ?
                   AND public_key = ?"""

        public_key = identity_key.get_public_key().serialize()
        result = self._con.execute(query, (recipient_id, public_key)).fetchone()

        return result is not None

    def delete_identity(self, recipient_id: str, identity_key: IdentityKey) -> None:
        query = """DELETE FROM identities
                   WHERE recipient_id = ? AND public_key = ?"""
        public_key = identity_key.get_public_key().serialize()
        self._con.execute(query, (recipient_id, public_key))
        self._con.commit()

    def is_trusted_identity(self, recipient_id: str, identity_key: IdentityKey) -> bool:
        return True

    def get_trust_for_identity(
        self, recipient_id: str, identity_key: IdentityKey
    ) -> OMEMOTrust | None:
        query = """SELECT trust FROM identities WHERE recipient_id = ?
                   AND public_key = ?"""
        public_key = identity_key.get_public_key().serialize()
        result = self._con.execute(query, (recipient_id, public_key)).fetchone()
        return result.trust if result is not None else None

    def get_fingerprints(self, jid: str):
        query = """SELECT recipient_id,
                          public_key as "public_key [pk]",
                          trust,
                          timestamp
                   FROM identities
                   WHERE recipient_id = ? ORDER BY trust ASC"""
        return self._con.execute(query, (jid,)).fetchall()

    def get_muc_fingerprints(self, jids: list[str]):
        query = """
            SELECT recipient_id,
                   public_key as "public_key [pk]",
                   trust,
                   timestamp
            FROM identities
            WHERE recipient_id IN ({}) ORDER BY trust ASC
            """.format(", ".join(["?"] * len(jids)))

        return self._con.execute(query, jids).fetchall()

    def get_default_trust(self, jid: str) -> OMEMOTrust:
        if not self._is_blind_trust_enabled():
            return OMEMOTrust.UNDECIDED

        query = """SELECT * FROM identities
                   WHERE recipient_id = ? AND trust IN (0, 1)"""
        result = self._con.execute(query, (jid,)).fetchone()
        if result is None:
            return OMEMOTrust.BLIND
        return OMEMOTrust.UNDECIDED

    def set_trust(
        self, recipient_id: str, identity_key: IdentityKey, trust: OMEMOTrust
    ) -> None:
        query = """UPDATE identities SET trust = ? WHERE public_key = ?
                   AND recipient_id = ?"""
        public_key = identity_key.get_public_key().serialize()
        self._con.execute(query, (trust, public_key, recipient_id))
        self._con.commit()

    def is_trusted(self, recipient_id: str, device_id: int) -> bool:
        record = self.load_session(recipient_id, device_id)
        if record.is_fresh():
            return False
        identity_key = record.get_session_state().get_remote_identity_key()
        return self.get_trust_for_identity(recipient_id, identity_key) in (
            OMEMOTrust.VERIFIED,
            OMEMOTrust.BLIND,
        )

    def get_identity_last_seen(
        self, recipient_id: str, identity_key: IdentityKey
    ) -> int | None:
        serialized = identity_key.get_public_key().serialize()
        query = """SELECT timestamp FROM identities
                   WHERE recipient_id = ? AND public_key = ?"""
        result = self._con.execute(query, (recipient_id, serialized)).fetchone()
        return result.timestamp if result is not None else None

    def set_identity_last_seen(
        self, recipient_id: str, identity_key: IdentityKey
    ) -> None:
        timestamp = int(time.time())
        serialized = identity_key.get_public_key().serialize()
        self._log.info("Set last seen for %s %s", recipient_id, timestamp)
        query = """UPDATE identities SET timestamp = ?
                   WHERE recipient_id = ? AND public_key = ?"""
        self._con.execute(query, (timestamp, recipient_id, serialized))
        self._con.commit()

    def get_unacknowledged_count(self, recipient_id: str, device_id: int) -> int:
        record = self.load_session(recipient_id, device_id)
        if record.is_fresh():
            return 0
        state = record.get_session_state()
        return state.get_sender_chain_key().get_index()

    def needs_init(self) -> bool:
        try:
            self.get_our_device_id()
        except AssertionError:
            return True
        return False



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0163: Personal Eventing Protocol

from __future__ import annotations

from nbxmpp.structs import DiscoInfo

from gajim.common import types
from gajim.common.modules.base import BaseModule


class PEP(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        for identity in info.identities:
            if identity.category == 'pubsub':
                if identity.type == 'pep':
                    self._log.info('Discovered PEP support: %s', info.jid)
                    self.supported = True



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common.i18n import _
from gajim.common.util.preview import GeoPreview

from gajim.gtk.menus import get_preview_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import GajimPopover


@Gtk.Template.from_string(string=get_ui_string("preview/geo.ui"))
class GeoPreviewWidget(Gtk.Box, SignalManager):
    __gtype_name__ = "GeoPreviewWidget"

    _image_button: Gtk.Button = Gtk.Template.Child()
    _location_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, preview: GeoPreview) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._uri = preview.uri
        self._location_label.set_text(preview.text)
        self._image_button.set_tooltip_text(_("Location at %s") % preview.text)
        self._image_button.set_action_target_value(GLib.Variant("s", preview.uri))

        self._menu_popover = GajimPopover(None)
        self.append(self._menu_popover)

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_preview_clicked)
        self.add_controller(gesture_secondary_click)

    def get_text(self) -> str:
        return self._uri

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._menu_popover
        Gtk.Box.do_unroot(self)

    def _on_preview_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        menu = get_preview_menu(self._uri)
        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x, y)
        self._menu_popover.popup()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
import time

import nbxmpp
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp.client import Client as NBXMPPClient
from nbxmpp.const import ConnectionType
from nbxmpp.const import StreamError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import modules
from gajim.common import passwords
from gajim.common.client_modules import ClientModules
from gajim.common.const import ClientState
from gajim.common.const import SimpleClientState
from gajim.common.events import AccountConnected
from gajim.common.events import AccountDisconnected
from gajim.common.events import MessageNotSent
from gajim.common.events import Notification
from gajim.common.events import PasswordRequired
from gajim.common.events import PlainConnection
from gajim.common.events import SignedIn
from gajim.common.events import StanzaReceived
from gajim.common.events import StanzaSent
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import determine_proxy
from gajim.common.helpers import get_account_proxy
from gajim.common.helpers import get_custom_host
from gajim.common.helpers import get_resource
from gajim.common.helpers import Observable
from gajim.common.helpers import warn_about_plain_connection
from gajim.common.i18n import _
from gajim.common.idle import IdleMonitorManager
from gajim.common.idle import Monitor
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.message import build_message_stanza
from gajim.common.modules.util import LogAdapter
from gajim.common.structs import OutgoingMessage
from gajim.common.util.standards import get_rfc5646_lang
from gajim.common.util.status import get_idle_status_message
from gajim.common.util.text import to_one_line

from gajim.gtk.util.window import open_window

log = logging.getLogger('gajim.c.client')


IgnoredTlsErrorsT = set[Gio.TlsCertificateFlags] | None


def call_counter(func: Any):
    def helper(self, restart: bool = False) -> Any:
        if restart:
            self._connect_machine_calls = 0
        self._connect_machine_calls += 1
        return func(self)
    return helper


class Client(Observable, ClientModules):
    def __init__(self, account: str) -> None:
        self._log = LogAdapter(log, {'account': account})
        Observable.__init__(self, self._log)
        ClientModules.__init__(self, account)
        self._client = None
        self._account = account
        self.name = account

        address = app.settings.get_account_setting(self._account, 'address')
        self._address = JID.from_string(address)

        self._connect_machine_calls = 0
        self.addressing_supported = False

        self.roster_supported = True

        self._state = ClientState.DISCONNECTED
        self._status_sync_on_resume = False
        self._status = 'online'
        self._status_message = ''
        self._idle_status = 'online'
        self._idle_status_enabled = True
        self._idle_status_message = ''

        self._reconnect = True
        self._reconnect_timer_source = None
        self._destroy_client = False
        self._remove_account = False
        self._connectivity = Gio.NetworkConnectivity.FULL

        self._destroyed = False

        self.available_transports = {}

        modules.register_modules(self)

        self._create_client()

        if Monitor.is_available():
            self._idle_handler_id = Monitor.connect('state-changed',
                                                    self._idle_state_changed)
            self._screensaver_handler_id = app.app.connect(
                'notify::screensaver-active', self._screensaver_state_changed)

        monitor = Gio.NetworkMonitor.get_default()
        self._network_monitor_id = monitor.connect(
            'notify::connectivity', self._network_status_changed)

    def _set_state(self, state: ClientState) -> None:
        self._log.info('State: %s', repr(state))
        self._state = state

    @property
    def state(self) -> ClientState:
        return self._state

    @property
    def account(self) -> str:
        return self._account

    @property
    def status(self) -> str:
        return self._status

    @property
    def status_message(self) -> str:
        if self._idle_status_active():
            return self._idle_status_message
        return self._status_message

    @property
    def certificate(self):
        return self._client.peer_certificate[0]

    @property
    def features(self):
        return self._client.features

    @property
    def local_address(self) -> str | None:
        address = self._client.local_address
        if address is not None:
            return address.to_string().split(':')[0]
        return None

    def is_destroyed(self) -> bool:
        return self._destroyed

    def set_remove_account(self, value: bool) -> None:
        # Used by the RemoveAccount Assistant to make the Client
        # not react to any stream errors that happen while the
        # account is removed by the server and the connection is killed
        self._remove_account = value

    def _create_client(self) -> None:
        self._log.info('Create new nbxmpp client')

        if self._client is not None:
            self._client.destroy()
            self._destroy_client = False

        self._client = NBXMPPClient(log_context=self._account)
        self.connection = self._client
        self._client.set_lang(get_rfc5646_lang())
        self._client.set_domain(self._address.domain)
        self._client.set_username(self._address.localpart)
        self._client.set_resource(get_resource(self._account))
        self._client.set_supported_fallback_ns([Namespace.REPLY])

        self._client.subscribe('resume-failed', self._on_resume_failed)
        self._client.subscribe('resume-successful', self._on_resume_successful)
        self._client.subscribe('disconnected', self._on_disconnected)
        self._client.subscribe('connection-failed', self._on_connection_failed)
        self._client.subscribe('connected', self._on_connected)

        self._client.subscribe('stanza-sent', self._on_stanza_sent)
        self._client.subscribe('stanza-received', self._on_stanza_received)

        for handler in modules.get_handlers(self):
            self._client.register_handler(handler)

    def _on_resume_failed(self,
                          _client: NBXMPPClient,
                          _signal_name: str) -> None:

        self._log.info('Resume failed')
        self.notify('resume-failed')

    def _on_resume_successful(self,
                              _client: NBXMPPClient,
                              _signal_name: str) -> None:

        self._set_state(ClientState.CONNECTED)
        self._set_client_available()

        if self._status_sync_on_resume:
            self._status_sync_on_resume = False
            self.update_presence()

        self.notify('state-changed', SimpleClientState.CONNECTED)
        self.notify('resume-successful')

    def _set_client_available(self) -> None:
        self._set_state(ClientState.AVAILABLE)
        app.ged.raise_event(AccountConnected(account=self._account))

        if not app.settings.get_account_setting(self._account, 'autojoin_sync'):
            self.join_mucs()

    def _network_status_changed(self,
                                monitor: Gio.NetworkMonitor,
                                _network_available: bool
                                ) -> None:

        if monitor.get_connectivity() == Gio.NetworkConnectivity.FULL:
            return

        reachable = self.remote_is_reachable()
        self._log.info('Network status changed, reachable: %s', reachable)

        if reachable:
            return

        if (self._state.is_connected or self._state.is_available):
            self.disconnect(gracefully=False, reconnect=True)

    def remote_is_reachable(self) -> bool:
        if self._client is None:
            return True

        address = self._client.remote_address
        if address is None:
            # Address is None when websocket is used
            self._log.info(
                "Unable to determine if host is reachable, remote address unknown")
            return True

        monitor = Gio.NetworkMonitor.get_default()
        try:
            address, port = address.rsplit(":", maxsplit=1)
            return monitor.can_reach(
                Gio.InetSocketAddress.new_from_string(address, int(port))
            )
        except GLib.Error as error:
            quark = GLib.quark_try_string('g-io-error-quark')
            if error.matches(quark, Gio.IOErrorEnum.HOST_UNREACHABLE):
                return False

            log.exception("Unable to determine if host is reachable")
            return True

        except Exception:
            log.exception("Unable to determine if host is reachable")
            return True

    def disconnect(self,
                   gracefully: bool,
                   reconnect: bool,
                   destroy_client: bool = False) -> None:

        self._reconnect = reconnect
        self._destroy_client = destroy_client

        if self._state.is_reconnect_scheduled:
            self._abort_reconnect()
            if reconnect:
                self.connect()
            return

        if self._state.is_disconnected:
            if destroy_client:
                self._create_client()
            if reconnect:
                self.connect()
            return

        if self._state.is_disconnecting:
            self._log.warning('Disconnect already in progress')
            return

        self._set_state(ClientState.DISCONNECTING)

        self._log.info('Starting to disconnect')
        self._client.disconnect(immediate=not gracefully)

    def _on_disconnected(self,
                         _client: NBXMPPClient,
                         _signal_name: str) -> None:

        self._log.info('Disconnect')
        self._set_state(ClientState.DISCONNECTED)

        domain, error, text = self._client.get_error()

        if self._remove_account:
            # Account was removed via RemoveAccount Assistant.
            self._reconnect = False

        elif domain == StreamError.BAD_CERTIFICATE:
            self._reconnect = False
            self._destroy_client = True

            cert, errors = self._client.peer_certificate

            open_window('SSLErrorDialog',
                        account=self._account,
                        client=self,
                        cert=cert,
                        ignored_errors=set(self._client.ignored_tls_errors),
                        error=errors.pop())

        elif domain in (StreamError.STREAM, StreamError.BIND):
            if error == 'conflict':
                # Reset resource
                app.settings.set_account_setting(self._account,
                                                 'resource',
                                                 'gajim.$rand')
            if error == 'system-shutdown':
                account_label = app.settings.get_account_setting(
                    self._account, "account_label"
                )
                account_text = _("Account: %s") % account_label
                shutdown_text = text or _("The server was shut down.")

                app.ged.raise_event(
                    Notification(
                        context_id="",
                        account=self._account,
                        type="server-shutdown",
                        title=_("Server Shutdown"),
                        text=f"{shutdown_text}\n{account_text}",
                    )
                )

        elif domain == StreamError.SASL:
            self._reconnect = False
            self._destroy_client = True

            if error in ('not-authorized', 'no-password'):
                def _on_password() -> None:
                    self.connect()

                app.ged.raise_event(PasswordRequired(client=self,
                                                     on_password=_on_password))

            app.ged.raise_event(
                Notification(context_id="",
                             account=self._account,
                             type='connection-failed',
                             title=_('Authentication failed'),
                             text=text or error))

        if self._reconnect:
            # Save resumeable information because in
            # _after_disconnect() we destroy the client object
            stream_is_resumeable = self._client.resumeable
            self._after_disconnect()
            self._schedule_reconnect()
            if not stream_is_resumeable:
                self.notify('state-changed', SimpleClientState.DISCONNECTED)
            self.notify('state-changed', SimpleClientState.RESUME_IN_PROGRESS)

        else:
            self._after_disconnect()
            self.notify('state-changed', SimpleClientState.DISCONNECTED)

    def _after_disconnect(self) -> None:
        self._disable_reconnect_timer()

        self.get_module('Bytestream').remove_all_transfers()

        if self._destroy_client:
            self._create_client()

        app.ged.raise_event(AccountDisconnected(account=self._account))

    def _on_connection_failed(self,
                              _client: NBXMPPClient,
                              _signal_name: str) -> None:
        self._schedule_reconnect()

    def _on_connected(self,
                      _client: NBXMPPClient,
                      _signal_name: str) -> None:

        self._set_state(ClientState.CONNECTED)
        self.get_module('Discovery').discover_server_info()
        self.get_module('Discovery').discover_account_info()
        self.get_module('Discovery').discover_server_items()

    def _on_stanza_sent(self,
                        _client: NBXMPPClient,
                        _signal_name: str,
                        stanza: Any) -> None:

        app.ged.raise_event(StanzaSent(account=self._account,
                                       stanza=stanza))

    def _on_stanza_received(self,
                            _client: NBXMPPClient,
                            _signal_name: str,
                            stanza: Any) -> None:

        app.ged.raise_event(StanzaReceived(account=self._account,
                                           stanza=stanza))

    def is_own_jid(self, jid: JID | str) -> bool:
        own_jid = self.get_own_jid()
        return own_jid.bare_match(jid)

    def get_own_contact(self) -> BareContact:
        jid = self.get_own_jid()
        contact = self.get_module("Contacts").get_contact(jid.bare)
        assert isinstance(contact, BareContact)
        return contact

    def get_own_jid(self) -> JID:
        '''
        Return the last full JID we received on a bind event.
        In case we were never connected it returns the bare JID from config.
        '''
        if self._client is not None:
            jid = self._client.get_bound_jid()
            if jid is not None:
                return jid

        # This returns the bare jid
        return nbxmpp.JID.from_string(app.get_jid_from_account(self._account))

    def get_bound_jid(self) -> JID:
        assert self._client is not None
        jid = self._client.get_bound_jid()
        assert jid is not None
        return jid

    def change_status(self, show: str, message: str) -> None:
        if not message:
            message = ''

        self._idle_status_enabled = show == 'online'
        self._status_message = message

        if show != 'offline':
            self._status = show

            app.settings.set_account_setting(
                self._account,
                'last_status',
                show)
            app.settings.set_account_setting(
                self._account,
                'last_status_msg',
                to_one_line(message))

        if self._state.is_disconnecting:
            self._log.warning("Can't change status while disconnect is in progress")
            return

        if self._state.is_disconnected:
            if show == 'offline':
                return

            self.connect()
            return

        if self._state.is_connecting:
            if show == 'offline':
                self.disconnect(gracefully=False,
                                reconnect=False,
                                destroy_client=True)
            return

        if self._state.is_reconnect_scheduled:
            if show == 'offline':
                self._destroy_client = True
                self._abort_reconnect()
            else:
                self.connect()
            return

        # We are connected
        if show == 'offline':
            self.get_module('UserTune').set_tune(None)
            self.get_module('UserLocation').set_location(None)
            presence = self.get_module('Presence').get_presence(
                typ='unavailable',
                status=message,
                caps=False)

            self.send_stanza(presence)
            self.disconnect(gracefully=True,
                            reconnect=False,
                            destroy_client=True)
            return

        self.update_presence()

    def update_presence(self, include_muc: bool = True) -> None:
        status, message, idle = self.get_presence_state()
        self.get_module('Presence').send_presence(
            show=status,
            status=message,
            idle_time=idle)

        if include_muc:
            self.get_module('MUC').update_presence()

    @call_counter
    def connect_machine(self) -> None:
        self._log.info('Connect machine state: %s', self._connect_machine_calls)
        if self._connect_machine_calls == 1:
            self.get_module('Roster').request_roster()
        elif self._connect_machine_calls == 2:
            self._finish_connect()

    def _finish_connect(self) -> None:
        self._status_sync_on_resume = False
        self._set_client_available()

        # We did not resume the stream, so we are not joined any MUCs
        self.update_presence(include_muc=False)

        self.get_module('Bookmarks').request_bookmarks()
        self.get_module('SoftwareVersion').set_enabled(True)
        self.get_module('LastActivity').set_enabled(True)
        self.get_module('EntityTime').set_enabled(True)
        self.get_module('Annotations').request_annotations()
        self.get_module('Blocking').get_blocking_list()
        self.get_module('VCard4').subscribe_to_node()

        if app.settings.get_account_setting(self._account, 'publish_tune'):
            self.get_module('UserTune').set_enabled(True)

        self.notify('state-changed', SimpleClientState.CONNECTED)

        app.ged.raise_event(SignedIn(account=self._account, conn=self))
        modules.send_stored_publish(self._account)

    def send_stanza(self, stanza: Any) -> None:
        '''
        Send a stanza untouched
        '''
        return self._client.send_stanza(stanza)

    def send_message(self, message: OutgoingMessage) -> None:
        if not self._state.is_available:
            self._log.warning('Trying to send message while offline')
            return

        stanza = build_message_stanza(message, self.get_own_jid())
        message.set_stanza(stanza)

        method = message.contact.settings.get('encryption')
        if not method:
            self._send_message(message)
            return

        if method == 'OMEMO':
            try:
                self.get_module('OMEMO').encrypt_message(message)
            except Exception:
                self._log.exception('Error')
                text = message.get_text(with_fallback=False)
                assert text is not None
                app.ged.raise_event(
                    MessageNotSent(
                        client=self._client,
                        jid=str(message.contact.jid),
                        message=text,
                        error=_('Encryption error'),
                        time=time.time()))
                return

            self._send_message(message)
            return

        # TODO: Make extension point return encrypted message
        extension = 'encrypt'
        if message.is_groupchat:
            extension = 'gc_encrypt'
        app.plugin_manager.extension_point(extension + method,
                                           self,
                                           message,
                                           self._send_message)

    def _send_message(self, message: OutgoingMessage) -> None:
        self.send_stanza(message.get_stanza())
        self.get_module('Message').store_message(message)

    def connect(
        self,
        ignored_tls_errors: IgnoredTlsErrorsT = None
    ) -> None:

        self._log.info('Connect')

        if self._state not in (ClientState.DISCONNECTED,
                               ClientState.RECONNECT_SCHEDULED):
            # Do not try to reco while we are already trying
            return

        assert self._client is not None
        custom_host = get_custom_host(self._account)
        if custom_host is not None:
            self._client.set_custom_host(*custom_host)

        gssapi = app.settings.get_account_setting(self._account,
                                                  'enable_gssapi')
        if gssapi:
            self._client.set_mechs(['GSSAPI'])

        anonymous = app.settings.get_account_setting(self._account,
                                                     'anonymous_auth')
        if anonymous:
            self._client.set_mechs(['ANONYMOUS'])

        if app.settings.get_account_setting(self._account,
                                            'use_plain_connection'):
            self._client.set_connection_types([ConnectionType.PLAIN])

        proxy = get_account_proxy(self._account)
        if proxy is not None:
            self._client.set_proxy(proxy)

        password = passwords.get_password(self._account)
        self._client.set_password(password)

        self._client.set_accepted_certificates(
            app.cert_store.get_certificates())
        self._client.set_ignored_tls_errors(ignored_tls_errors)

        self._reconnect = True
        self._disable_reconnect_timer()
        self._set_state(ClientState.CONNECTING)
        self.notify('state-changed', SimpleClientState.CONNECTING)

        if warn_about_plain_connection(self._account,
                                       self._client.connection_types):
            app.ged.raise_event(PlainConnection(account=self._account,
                                                connect=self._client.connect,
                                                abort=self._abort_reconnect))
            return

        def _on_host_meta_response(obj: FileTransfer) -> None:
            if self._client is None or self._state != ClientState.CONNECTING:
                # State was changed in the meantime
                return

            try:
                result = obj.get_result()
            except Exception as error:
                self._log.warning('Error while requesting host-meta data: %s', error)
            else:
                self._log.info("Received host meta data with length: %s",
                               len(result.content))
                self._client.set_host_meta_data(result.content)

            self._client.connect()

        app.ftm.http_request(
            'GET',
            f'https://{self._address.domain}/.well-known/host-meta',
            proxy=determine_proxy(self._account),
            timeout=3,
            callback=_on_host_meta_response,
        )

    def _schedule_reconnect(self) -> None:
        self._set_state(ClientState.RECONNECT_SCHEDULED)
        self._log.info('Reconnect in 3s')
        self._reconnect_timer_source = GLib.timeout_add_seconds(
            3, self.connect)

    def _abort_reconnect(self) -> None:
        self._set_state(ClientState.DISCONNECTED)
        self._disable_reconnect_timer()

        if self._destroy_client:
            self._create_client()

        self.notify('state-changed', SimpleClientState.DISCONNECTED)

    def _disable_reconnect_timer(self) -> None:
        if self._reconnect_timer_source is not None:
            GLib.source_remove(self._reconnect_timer_source)
            self._reconnect_timer_source = None

    def _idle_state_changed(self, monitor: IdleMonitorManager) -> None:
        state = monitor.state.value

        if monitor.is_awake():
            self._idle_status = state
            self._idle_status_message = ''
            self._update_status()
            return

        if not app.settings.get(f'auto{state}'):
            return

        if ((state in ('away', 'xa') and self._status == 'online') or
                (state == 'xa' and self._idle_status == 'away')):

            self._idle_status = state
            self._idle_status_message = get_idle_status_message(
                state, self._status_message)
            self._update_status()

    def _update_status(self) -> None:
        if not self._idle_status_enabled:
            return

        self._status = self._idle_status
        if self._state.is_available:
            self.update_presence()
        else:
            self._status_sync_on_resume = True

    def _idle_status_active(self) -> bool:
        if not Monitor.is_available():
            return False

        if not self._idle_status_enabled:
            return False

        return self._idle_status != 'online'

    def get_presence_state(self) -> tuple[str, str, bool]:
        if self._idle_status_active():
            return self._idle_status, self._idle_status_message, True
        return self._status, self._status_message, False

    @staticmethod
    def _screensaver_state_changed(application: Gtk.Application,
                                   _param: GObject.ParamSpec) -> None:
        active = application.get_property('screensaver-active')
        Monitor.set_extended_away(active)

    def join_mucs(self) -> None:
        '''Only used when autojoin_sync is False'''
        self._log.info('Joining all MUCs in all workspaces')
        for workspace_id in app.settings.get_workspaces():
            for chat in app.settings.get_workspace_setting(
                    workspace_id, 'chats'):
                if chat['account'] != self._account:
                    continue
                if chat['type'] == 'groupchat':
                    self.get_module('MUC').join(chat['jid'])

    def cleanup(self) -> None:
        self.disconnect_signals()
        self._destroyed = True
        if Monitor.is_available():
            Monitor.disconnect(self._idle_handler_id)
            app.app.disconnect(self._screensaver_handler_id)

        monitor = Gio.NetworkMonitor.get_default()
        monitor.disconnect(self._network_monitor_id)

        if self._client is not None:
            self._client.destroy()
        modules.unregister_modules(self)

    def quit(self, kill_core: bool) -> None:
        if kill_core and self._state in (ClientState.CONNECTING,
                                         ClientState.CONNECTED,
                                         ClientState.AVAILABLE):
            self.disconnect(gracefully=True, reconnect=False)



#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import logging
from pathlib import Path

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.modules.user_avatar import Avatar
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.modules.contacts import BareContact

from gajim.gtk.avatar import clip_circle
from gajim.gtk.avatar_selector import AvatarSelector
from gajim.gtk.builder import get_builder
from gajim.gtk.filechoosers import AvatarFileChooserButton
from gajim.gtk.util.misc import convert_surface_to_texture
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.util.misc import scroll_to
from gajim.gtk.vcard_grid import VCardGrid
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.profile")

MENU_DICT = {
    "fn": p_("Profile", "Full Name"),
    "bday": _("Birthday"),
    "gender": p_("Profile", "Gender"),
    "pronouns": _("Pronouns"),
    "adr": p_("Profile", "Address"),
    "email": _("Email"),
    "impp": p_("Profile", "IM Address"),
    "tel": _("Phone No."),
    "org": p_("Profile", "Organisation"),
    "title": p_("Profile", "Title"),
    "role": p_("Profile", "Role"),
    "url": _("URL"),
    "key": p_("Profile", "Public Encryption Key"),
    "note": p_("Profile", "Note"),
    "tz": p_("Profile", "Timezone"),
}


class ProfileWindow(GajimAppWindow):
    def __init__(self, account: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="ProfileWindow",
            title=_("Profile"),
            default_width=700,
            default_height=600,
            add_window_padding=True,
            header_bar=True,
        )

        self.account = account
        self._jid = app.get_jid_from_account(account)
        self._running_tasks: list[Task] = []
        self._destroyed = False

        self._client = app.get_client(self.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._contact = self._client.get_module("Contacts").get_contact(self._jid)

        self._ui = get_builder("profile.ui")

        menu = Gio.Menu()
        for action, label in MENU_DICT.items():
            menu.append(label, f"win.add-{action.lower()}")

        self._ui.add_entry_button.set_menu_model(menu)
        self._add_actions()

        self._connect(self._ui.back_button, "clicked", self._on_back_clicked)
        self._connect(self._ui.avatar_update_button, "clicked", self._on_update_avatar)
        self._connect(self._ui.edit_button, "clicked", self._on_edit_clicked)
        self._connect(self._ui.avatar_cancel, "clicked", self._on_cancel_update_avatar)
        self._connect(self._ui.save_button, "clicked", self._on_save_clicked)
        self._connect(self._ui.cancel_button, "clicked", self._on_cancel_clicked)
        self._connect(
            self._ui.avatar_nick_access, "notify::active", self._access_switch_toggled
        )
        self._connect(self._ui.remove_avatar_button, "clicked", self._on_remove_avatar)
        self._connect(
            self._ui.vcard_access, "notify::active", self._access_switch_toggled
        )

        self._avatar_edit_button = AvatarFileChooserButton(
            tooltip=_("Change your profile picture"),
            icon_name="lucide-square-pen-symbolic",
        )
        self._avatar_edit_button.add_css_class("opaque")
        self._connect(self._avatar_edit_button, "path-picked", self._on_edit_avatar)

        self._avatar_edit_button.set_halign(Gtk.Align.END)
        self._avatar_edit_button.set_valign(Gtk.Align.END)
        self._avatar_edit_button.set_visible(False)
        self._ui.avatar_overlay.add_overlay(self._avatar_edit_button)

        self._avatar_selector: AvatarSelector | None = None
        self._current_avatar: Gdk.Texture | None = None
        self._current_vcard: VCard | None = None
        self._avatar_nick_public: bool | None = None

        # False  - no change to avatar
        # None   - we want to delete the avatar
        # Avatar - upload new avatar
        self._new_avatar: None | bool | Avatar = False

        self._ui.avatar_image.set_pixel_size(AvatarSize.VCARD)
        self._ui.nickname_label.set_text(app.nicks[account])
        self._ui.nickname_entry.set_text(app.nicks[account])

        self._vcard_grid = VCardGrid(self.account)
        self._ui.profile_box.append(self._vcard_grid)

        self.set_child(self._ui.profile_stack)

        self._load_avatar()

        client = app.get_client(account)
        own_jid = client.get_own_jid().new_as_bare()
        client.get_module("VCard4").request_vcard(
            own_jid, callback=self._on_vcard_received
        )

        client.get_module("PubSub").get_access_model(
            Namespace.VCARD4_PUBSUB,
            callback=self._on_access_model_received,
            user_data=Namespace.VCARD4_PUBSUB,
        )

        client.get_module("PubSub").get_access_model(
            Namespace.AVATAR_METADATA,
            callback=self._on_access_model_received,
            user_data=Namespace.AVATAR_METADATA,
        )

        client.get_module("PubSub").get_access_model(
            Namespace.AVATAR_DATA,
            callback=self._on_access_model_received,
            user_data=Namespace.AVATAR_DATA,
        )

        client.get_module("PubSub").get_access_model(
            Namespace.NICK,
            callback=self._on_access_model_received,
            user_data=Namespace.NICK,
        )

    def _cleanup(self) -> None:
        self._destroyed = True
        del self._avatar_edit_button
        self._running_tasks.clear()
        self._avatar_selector = None
        self._client.disconnect_all_from_obj(self)

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._ui.save_button.set_sensitive(state.is_connected)
        self._ui.save_button.set_tooltip_text(
            _("You are offline") if not state.is_connected else ""
        )

    @ensure_not_destroyed
    def _on_access_model_received(self, task: Task) -> None:
        namespace = task.get_user_data()

        try:
            result = task.finish()
        except StanzaError as error:
            log.warning("Unable to get access model for %s: %s", namespace, error)
            return

        access_model = result == "open"

        if namespace == Namespace.VCARD4_PUBSUB:
            self._set_vcard_access_switch(access_model)
        else:
            if self._avatar_nick_public is None:
                self._avatar_nick_public = access_model
            else:
                self._avatar_nick_public = self._avatar_nick_public or access_model
            assert self._avatar_nick_public is not None
            self._set_avatar_nick_access_switch(self._avatar_nick_public)

    @ensure_not_destroyed
    def _on_vcard_received(self, jid: JID, vcard: VCard):
        self._current_vcard = vcard

        self._load_avatar()
        self._vcard_grid.set_vcard(self._current_vcard.copy())
        self._show_profile_page()

    def _load_avatar(self) -> None:
        scale = self.get_scale_factor()
        assert isinstance(self._contact, BareContact)
        self._current_avatar = self._contact.get_avatar(
            AvatarSize.VCARD, scale, add_show=False
        )
        self._ui.avatar_image.set_from_paintable(self._current_avatar)
        self._ui.avatar_image.set_visible(True)

    def _add_actions(self) -> None:
        for action in MENU_DICT:
            action_name = f"add-{action.lower()}"
            act = Gio.SimpleAction.new(action_name, None)
            self._connect(act, "activate", self._on_action)
            self.add_action(act)

    def _on_action(self, action: Gio.SimpleAction, _param: GLib.Variant | None) -> None:
        name = action.get_name()
        key = name.split("-")[1]
        self._vcard_grid.add_new_property(key)
        GLib.idle_add(scroll_to, self._ui.scrolled, "bottom")

    def _on_edit_clicked(self, *args: Any) -> None:
        self._vcard_grid.set_editable(True)
        self._ui.edit_button.set_visible(False)

        self._ui.add_entry_button.set_visible(True)
        self._ui.cancel_button.set_visible(True)
        self._ui.save_button.set_visible(True)
        self._ui.remove_avatar_button.set_visible(True)
        self._avatar_edit_button.set_visible(True)
        self._ui.nickname_label.set_visible(False)
        self._ui.nickname_entry.set_visible(True)
        self._ui.privacy_button.set_visible(True)

    def _on_cancel_clicked(self, _button: Gtk.Button) -> None:
        self._vcard_grid.set_editable(False)
        self._ui.edit_button.set_visible(True)
        self._ui.add_entry_button.set_visible(False)
        self._ui.cancel_button.set_visible(False)
        self._ui.save_button.set_visible(False)
        self._ui.remove_avatar_button.set_visible(False)
        self._avatar_edit_button.set_visible(False)
        self._ui.privacy_button.set_visible(False)
        self._ui.nickname_label.set_visible(True)
        self._ui.nickname_label.set_text(app.nicks[self.account])
        self._ui.nickname_entry.set_visible(False)
        self._ui.nickname_entry.set_text(app.nicks[self.account])
        self._ui.avatar_image.set_from_paintable(self._current_avatar)
        self._vcard_grid.set_vcard(self._current_vcard.copy())  # pyright: ignore
        self._new_avatar = False

    def _on_back_clicked(self, *args: Any) -> None:
        self._show_profile_page()

    def _on_save_clicked(self, _button: Gtk.Button) -> None:
        self._ui.add_entry_button.set_visible(False)
        self._ui.cancel_button.set_visible(False)
        self._ui.save_button.set_visible(False)
        self._ui.edit_button.set_visible(True)
        self._ui.remove_avatar_button.set_visible(False)
        self._avatar_edit_button.set_visible(False)
        self._ui.privacy_button.set_visible(False)
        self._ui.nickname_label.set_visible(True)
        self._ui.nickname_entry.set_visible(False)

        # Switch page after setting other widget's visibility
        # This avoids a leak caused by switching stack pages
        self._ui.profile_stack.set_visible_child_name("spinner")

        self._vcard_grid.validate()
        self._vcard_grid.sort()

        vcard = self._vcard_grid.get_vcard()
        self._current_vcard = vcard.copy()  # pyright: ignore

        client = app.get_client(self.account)
        task = client.get_module("VCard4").set_vcard(
            self._current_vcard,
            public=self._ui.vcard_access.get_active(),
            callback=self._on_set_vcard_result,
        )
        self._running_tasks.append(task)

        public = self._ui.avatar_nick_access.get_active()

        if self._new_avatar is False:
            if self._avatar_nick_public != public:
                client.get_module("UserAvatar").set_access_model(public)

        else:
            # Only update avatar if it changed
            task = client.get_module("UserAvatar").set_avatar(
                self._new_avatar, public=public, callback=self._on_set_avatar_result
            )
            self._running_tasks.append(task)

        self._avatar_nick_public = public

        nick = self._ui.nickname_entry.get_text()
        client.get_module("UserNickname").set_nickname(nick, public=public)

        if not nick:
            nick = app.get_default_nick(self.account)
        app.nicks[self.account] = nick

    @ensure_not_destroyed
    def _on_set_avatar_result(self, task: Task) -> None:
        self._running_tasks.remove(task)
        try:
            task.finish()
        except StanzaError as error:
            if self._new_avatar is None:
                # Trying to remove the avatar but the node does not exist
                if error.condition == "item-not-found":
                    return

            title = _("Error while uploading avatar")
            text = error.get_text()

            if (
                error.condition == "not-acceptable"
                and error.app_condition == "payload-too-big"
            ):
                text = _("Avatar file size too big")

            self._ui.avatar_image.set_from_paintable(self._current_avatar)
            self._new_avatar = False

            self._show_error_page(title, text)
            return

        self._client.update_presence(include_muc=True)
        if not self._running_tasks:
            self._show_profile_page()

    @ensure_not_destroyed
    def _on_set_vcard_result(self, task: Task) -> None:
        self._running_tasks.remove(task)
        try:
            task.finish()
        except StanzaError as error:
            log.error("Could not publish VCard: %s", error)
            self._show_error_page(_("Unable to save profile"), error.get_text())
            return

        self._vcard_grid.set_editable(False)
        if not self._running_tasks:
            self._show_profile_page()

    def _on_remove_avatar(self, _button: Gtk.Button) -> None:
        scale = self.get_scale_factor()
        assert isinstance(self._contact, BareContact)
        texture = self._contact.get_avatar(
            AvatarSize.VCARD, scale, add_show=False, default=True
        )

        self._ui.avatar_image.set_from_paintable(texture)
        self._ui.remove_avatar_button.set_visible(False)
        self._new_avatar = None

    def _on_edit_avatar(
        self, _button: AvatarFileChooserButton, paths: list[Path]
    ) -> None:
        if self._avatar_selector is None:
            self._avatar_selector = AvatarSelector()
            self._ui.avatar_selector_box.prepend(self._avatar_selector)

        self._avatar_selector.prepare_crop_area(str(paths[0]))
        self._ui.avatar_update_button.set_sensitive(
            self._avatar_selector.get_prepared()
        )
        self._ui.profile_stack.set_visible_child_name("avatar_selector")

    def _on_cancel_update_avatar(self, _button: Gtk.Button) -> None:
        self._show_profile_page()

    def _on_update_avatar(self, _button: Gtk.Button) -> None:
        error_title = _("Error while processing image")
        error_text = _("Failed to generate avatar.")

        assert self._avatar_selector
        success, data, width, height = self._avatar_selector.get_avatar_bytes()
        if not success:
            self._show_error_page(error_title, error_text)
            return

        assert data
        assert height
        assert width
        sha = app.app.avatar_storage.save_avatar(data)
        if sha is None:
            self._show_error_page(error_title, error_text)
            return

        self._new_avatar = Avatar()
        self._new_avatar.add_image_source(data, "image/png", height, width)

        scale = self.get_scale_factor()
        surface = app.app.avatar_storage.surface_from_filename(
            sha, AvatarSize.VCARD, scale
        )
        if surface is None:
            return

        self._ui.avatar_image.set_from_paintable(
            convert_surface_to_texture(clip_circle(surface))
        )
        self._ui.remove_avatar_button.set_visible(True)
        self._show_profile_page()

    def _set_vcard_access_switch(self, state: bool) -> None:
        self._ui.vcard_access.set_active(state)
        self._ui.vcard_access_label.set_text(_("Everyone") if state else _("Contacts"))

    def _set_avatar_nick_access_switch(self, state: bool) -> None:
        self._ui.avatar_nick_access.set_active(state)
        self._ui.avatar_nick_access_label.set_text(
            _("Everyone") if state else _("Contacts")
        )

    def _access_switch_toggled(self, *args: Any) -> None:
        state = self._ui.vcard_access.get_active()
        self._set_vcard_access_switch(state)

        state = self._ui.avatar_nick_access.get_active()
        self._set_avatar_nick_access_switch(state)

    def _show_error_page(self, title: str, text: str) -> None:
        self._ui.error_title_label.set_text(title)
        self._ui.error_label.set_text(text)
        self._ui.profile_stack.set_visible_child_name("error")

    def _show_profile_page(self) -> None:
        self._running_tasks.clear()
        self._ui.profile_stack.set_visible_child_name("profile")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import logging
from datetime import datetime
from pathlib import Path

from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import configpaths
from gajim.common.helpers import make_path_from_jid
from gajim.common.i18n import _
from gajim.common.modules.contacts import ResourceContact
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.builder import get_builder
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.filechoosers import FileChooserButton

log = logging.getLogger("gajim.gtk.history_export")


class HistoryExport(Assistant):
    def __init__(self, account: str | None = None, jid: JID | None = None) -> None:
        Assistant.__init__(self)

        self.account = account
        self.jid = jid

        self.add_button("back", _("Back"))
        self.add_button("close", _("Close"))
        self.add_button(
            "export", _("Export"), complete=True, css_class="suggested-action"
        )

        self.set_button_visible_func(self._visible_func)

        self.add_pages({"start": ExportSettings(account, jid)})

        progress_page = self.add_default_page("progress")
        progress_page.set_title(_("Exporting History..."))
        progress_page.set_text(_("Exporting your messages..."))

        success_page = self.add_default_page("success")
        success_page.set_title(_("Export Finished"))
        success_page.set_text(_("Your messages have been exported successfully"))

        error_page = self.add_default_page("error")
        error_page.set_title(_("Error while Exporting"))
        error_page.set_text(_("An error occurred while exporting your messages"))

        self._connect(self, "button-clicked", self._on_button_clicked)

    @overload
    def get_page(self, name: Literal["error"]) -> AssistantErrorPage: ...

    @overload
    def get_page(self, name: Literal["start"]) -> ExportSettings: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    @staticmethod
    def _visible_func(_assistant: Assistant, page_name: str) -> list[str]:
        if page_name == "start":
            return ["close", "export"]

        if page_name == "progress":
            return []

        if page_name == "success":
            return ["back", "close"]

        if page_name == "error":
            return ["back", "close"]
        raise ValueError(f"page {page_name} unknown")

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        if button_name == "export":
            self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)
            self._on_export()

        elif button_name == "back":
            self.show_page("start", Gtk.StackTransitionType.SLIDE_RIGHT)

        elif button_name == "close":
            self.close()

    def _on_export(self) -> None:
        start_page = self.get_page("start")
        account, jid, directory = start_page.get_export_settings()

        current_time = datetime.now()
        time_str = current_time.strftime("%Y-%m-%d-%H-%M-%S")
        export_dir = directory / f"export_{time_str}"

        if jid is None:
            rows = app.storage.archive.get_conversation_jids(account)
            jids = [jid for jid, _m_type in rows]
        else:
            jids = [jid]

        for jid in jids:
            messages = app.storage.archive.get_messages_for_export(account, jid)
            if not messages:
                continue

            file_path = make_path_from_jid(export_dir, jid)
            try:
                file_path.mkdir(parents=True, exist_ok=True)
            except OSError as err:
                self.get_page("error").set_text(
                    _(
                        "An error occurred while trying to create a "
                        "file at %(path)s: %(error)s"
                    )
                    % {"path": file_path, "error": str(err)}
                )
                self.show_page("error", Gtk.StackTransitionType.SLIDE_LEFT)
                return

            with open(file_path / "history.txt", "w", encoding="utf-8") as file:
                file.write(f"History for {jid}\n\n")
                for message in messages:
                    if message.call is not None:
                        continue

                    file.write(self._get_export_line(message))

        self.show_page("success", Gtk.StackTransitionType.SLIDE_LEFT)

    def _get_nickname(self, message: Message) -> str:
        if message.direction == ChatDirection.OUTGOING:
            return _("You")

        if message.type == MessageType.GROUPCHAT:
            if message.occupant is not None:
                if message.occupant.nickname is not None:
                    return message.occupant.nickname

            if message.resource is not None:
                return message.resource

            return _("Group Chat")

        return str(message.remote.jid)

    def _get_export_line(self, message: Message) -> str:
        name = self._get_nickname(message)
        timestamp = message.timestamp.astimezone().strftime("%Y-%m-%d %H:%M:%S")

        text = message.text
        if corrected_message := message.get_last_correction():
            text = corrected_message.text

        return f"{timestamp} {name}: {text or ''}\n"


class ExportSettings(AssistantPage):
    def __init__(self, account: str | None, jid: JID | None) -> None:
        AssistantPage.__init__(self)
        self._account = account
        self._jid = jid

        self._export_directory = configpaths.get("MY_DATA")

        self.title = _("Export Chat History")

        self._ui = get_builder("history_export.ui")
        self.append(self._ui.select_account_box)

        accounts_data: dict[str, str] = {}
        for account_data in app.get_enabled_accounts_with_labels():
            accounts_data[account_data[0]] = account_data[1]

        self._accounts_dropdown: GajimDropDown[str] = GajimDropDown(
            data=accounts_data, fixed_width=40
        )
        self._ui.settings_grid.attach(self._accounts_dropdown, 1, 0, 1, 1)
        self._connect(
            self._accounts_dropdown, "notify::selected", self._on_account_changed
        )

        self._chats_dropdown: GajimDropDown[str] = GajimDropDown(fixed_width=40)
        self._update_chat_dropdown()
        self._connect(self._chats_dropdown, "notify::selected", self._on_chat_changed)
        self._ui.settings_grid.attach(self._chats_dropdown, 1, 1, 1, 1)

        if self._account is not None:
            self._accounts_dropdown.select_key(self._account)

        if self._jid is not None:
            self._chats_dropdown.select_key(str(self._jid))

        file_chooser_button = FileChooserButton(
            path=self._export_directory,
            mode="folder-open",
            label=_("Choose History Export Directory"),
        )
        file_chooser_button.set_size_request(250, -1)
        self._connect(file_chooser_button, "path-picked", self._on_path_picked)
        self._ui.settings_grid.attach(file_chooser_button, 1, 2, 1, 1)

        self._set_complete()

    def _on_account_changed(self, dropdown: GajimDropDown[str], *args: Any) -> None:
        item = dropdown.get_selected_item()
        assert item is not None
        account = item.key
        self._account = account
        self._update_chat_dropdown()
        self._set_complete()

    def _on_chat_changed(self, dropdown: GajimDropDown[str], *args: Any) -> None:
        item = dropdown.get_selected_item()
        if item is None:
            return

        address = item.key
        if address:
            self._jid = JID.from_string(address)
        else:
            self._jid = None

    def _update_chat_dropdown(self) -> None:
        if self._account is None:
            self._chats_dropdown.set_data(None)
            return

        rows = app.storage.archive.get_conversation_jids(self._account)
        client = app.get_client(self._account)

        chats: dict[str, str] = {"": _("All Chats")}
        for jid, m_type in rows:
            contact = client.get_module("Contacts").get_contact(
                jid, groupchat=m_type != MessageType.CHAT
            )
            assert not isinstance(contact, ResourceContact)
            chats[str(jid)] = f"{contact.name} ({jid})"

        self._chats_dropdown.set_data(chats)

    def _set_complete(self) -> None:
        self.complete = bool(self._account is not None)
        self.update_page_complete()

    def _on_path_picked(
        self, _file_chooser_button: FileChooserButton, paths: list[Path]
    ) -> None:
        if not paths:
            return
        self._export_directory = paths[0]

    def get_export_settings(self) -> tuple[str, JID | None, Path]:
        assert self._account is not None
        assert self._export_directory is not None
        return self._account, self._jid, self._export_directory
